__author__ = 'bkeroack'

import deployment.salt_control
import deployment.sshconfig

def setup_new_server(datasvc, name):
    '''
    Do any initial server setup so it works for elita.

    Initially we just check if it's a Windows server and if so set up the git wrapper. At some point maybe we should
    verify that git is available (for all platforms) and whatever else is needed.
    '''
    ss = ServerSetup(datasvc)
    ss.new(name)

class ServerSetup:
    '''
    Initialization routines for new servers.
    '''
    def __init__(self, datasvc):
        '''
        @type datasvc: elita.models.DataService
        '''
        self.datasvc = datasvc

    def set_server_type(self, rc, name):
        os_type = rc.get_os_text(name)
        self.datasvc.serversvc.UpdateServer(name, {'server_type': os_type})
        return os_type

    def set_server_status(self, rc, name):
        if rc.sc.verify_connectivity(name):
            status = "ok"
        else:
            status = "not connectable (check salt)"

        self.datasvc.serversvc.UpdateServer(name, {'status': status})
        return status == "ok"

    def do_server_setup(self, rc, name, os_type):
        #create key/config dirs
        sshc = deployment.sshconfig.SSHController()
        sshc.create_key_dir(rc, [name])

        if os_type == "windows":
            # push git wrapper script. run it.
            # make sure git works
            wrapper_setup_path = "C:/git_wrapper_setup.ps1"
            rc.rm_file_if_exists([name], wrapper_setup_path)
            rc.push_files([name], {"salt://elita/files/win/git_wrapper_setup.ps1": wrapper_setup_path})
            rc.run_powershell_script([name], wrapper_setup_path)
            if 'not recognized as an internal or external command' in rc.run_shell_command([name], 'git')[name]:
                status = 'git not available'
            else:
                self.datasvc.jobsvc.NewJobData({'note': 'IMPORTANT!! For the Windows git wrapper script to take effect '
                                                        '(necessary for key/hostname management), you *must* either '
                                                        'reboot or restart salt-minion at a minimum! Any git operation '
                                                        'will fail until this happens.'})
                status = 'ok, initialized'
        else:
            status = 'ok, initialized'
        self.datasvc.serversvc.UpdateServer(name, {'status': status})

    def new(self, name):
        sc = deployment.salt_control.SaltController(self.datasvc)
        rc = deployment.salt_control.RemoteCommands(sc)

        ost = self.set_server_type(rc, name)
        if self.set_server_status(rc, name):
            self.do_server_setup(rc, name, ost)



__author__ = 'bkeroack'

import elita.deployment.deploy

import pprint
pp = pprint.PrettyPrinter(indent=4)

divisor = 2

rolling_groups = {
    "group0":
    {
        "servers": ["server0", "server1", "server2", "server3"],
        "gitdeploys": ["gd0", "gd1"]
    },
    "group1":
    {
        "servers": ["server4", "server5", "server6", "server7"],
        "gitdeploys": ["gd0", "gd2"]
    }
}

nonrolling_groups = {
        "group2":
        {
            "servers": ["server8", "server9"],
            "gitdeploys": ["gd0", "gd3"]
        }
}

ordered_groups = {
    "group3":
        {
            "servers": ["server10", "server11"],
            "gitdeploys": [["gd9"], ["gd4"]]
        },
    "group4":
        {
            "servers": ["server12", "server13"],
            "gitdeploys": [["gd9"], ["gd5"]]
        }
}

simulated_scorebig_rolling_groups = {
    "ConsumerWebApp": {
        "servers": ["consumer0", "consumer1"],
        "gitdeploys": [["Configs"], ["Consumer"]]
    },
    "ExtranetWebApp": {
        "servers": ["extranet0", "extranet1"],
        "gitdeploys": [["Configs"], ["Extranet"]]
    }
}

simulated_scorebig_nonrolling_groups = {
    "ServiceBusLowApplication": {
        "servers": ["sbl0", "sbl1"],
        "gitdeploys": [["Configs"], ["ServiceBusLow"]]
    },
    "ServiceBusHighApplication": {
        "servers": ["sbh0", "sbh1"],
        "gitdeploys": [["Configs"], ["ServiceBusHigh"]]
    }
}

def generate_server_batch_mapping(batches):
    server_batch_mapping = dict()
    for i, b in enumerate(batches):
        for s in b['servers']:
            for group in ordered_groups:
                if s in ordered_groups[group]['servers']:
                    for g in b['gitdeploys']:
                        if s in server_batch_mapping:
                            server_batch_mapping[s][g] = i
                        else:
                            server_batch_mapping[s] = {g: i}
    return server_batch_mapping

def test_rolling_batches():
    '''
    Test that rolling groups are split up into the requested batches
    '''

    batches = elita.deployment.deploy.BatchCompute.compute_rolling_batches(divisor, rolling_groups, None)

    assert len(batches) == 2
    assert all(["servers" in x and "gitdeploys" in x for x in batches])
    assert all([len(x['servers']) == 4 for x in batches])
    assert sorted(batches[0]['servers']) == sorted(['server0', 'server1', 'server4', 'server5'])
    assert sorted(batches[1]['servers']) == sorted(['server2', 'server3', 'server6', 'server7'])
    assert all([sorted(x['gitdeploys']) == sorted(['gd0', 'gd1', 'gd2']) for x in batches])
    assert not any([x['ordered_gitdeploy'] for x in batches])


def test_rolling_and_nonrolling_batches():
    '''
    Test that simultaneous rolling and non-rolling groups are split into appropriate batches
    '''

    batches = elita.deployment.deploy.BatchCompute.compute_rolling_batches(divisor, rolling_groups, nonrolling_groups)

    assert len(batches) == 2
    assert all(["servers" in x and "gitdeploys" in x for x in batches])
    assert len(batches[0]['servers']) == 6
    assert len(batches[1]['servers']) == 4
    assert sorted(batches[0]['servers']) == sorted(['server0', 'server1', 'server4', 'server5', 'server8', 'server9'])
    assert sorted(batches[1]['servers']) == sorted(['server2', 'server3', 'server6', 'server7'])
    assert sorted(batches[0]['gitdeploys']) == sorted(['gd0', 'gd1', 'gd2', 'gd3'])
    assert sorted(batches[1]['gitdeploys']) == sorted(['gd0', 'gd1', 'gd2'])
    assert not any([x['ordered_gitdeploy'] for x in batches])

def test_ordered_rolling_batches():
    '''
    Test that ordered rolling groups are split into batches respecting the ordering
    '''

    batches = elita.deployment.deploy.BatchCompute.compute_rolling_batches(divisor, ordered_groups, None)

    assert len(batches) == 4
    assert all(["servers" in x and "gitdeploys" in x for x in batches])
    assert all([len(x['servers']) == 2 for x in batches])
    assert all([batches[x]['gitdeploys'] == ['gd9'] for x in (0, 2)])
    assert all([sorted(batches[x]['gitdeploys']) == sorted(['gd4', 'gd5']) for x in (1, 3)])
    assert all([sorted(batches[x]['servers']) == sorted(['server12', 'server10']) for x in (0, 1)])
    assert all([sorted(batches[x]['servers']) == sorted(['server11', 'server13']) for x in (2, 3)])
    assert all([x['ordered_gitdeploy'] for x in batches[0::2]])
    assert not any([x['ordered_gitdeploy'] for x in batches[1::2]])


def test_ordered_and_unordered_rolling_batches():
    '''
    Test that simultaneous ordered and unordered rolling groups are split into batches respecting ordering
    '''

    batches = elita.deployment.deploy.BatchCompute.compute_rolling_batches(divisor, dict(ordered_groups,
                                                                                         **rolling_groups), None)

    assert len(batches) == 4
    assert all(["servers" in x and "gitdeploys" in x for x in batches])
    assert not any([x['ordered_gitdeploy'] for x in batches])

    server_batch_mapping = generate_server_batch_mapping(batches)
    for s in server_batch_mapping:
        if 'gd4' in server_batch_mapping[s]:
            assert server_batch_mapping[s]['gd9'] < server_batch_mapping[s]['gd4']
        elif 'gd5' in server_batch_mapping[s]:
            assert server_batch_mapping[s]['gd9'] < server_batch_mapping[s]['gd5']

def test_simulated_scorebig_mixed_groups():
    '''
    Test that ordered and unordered groups similar to ScoreBig setup are split into batches respecting ordering
    '''
    batches = elita.deployment.deploy.BatchCompute.compute_rolling_batches(divisor, simulated_scorebig_rolling_groups,
                                                                           simulated_scorebig_nonrolling_groups)

    assert len(batches) == 4
    assert all(["servers" in x and "gitdeploys" in x for x in batches])
    assert all([x['ordered_gitdeploy'] for x in batches[0::2]])
    assert not any([x['ordered_gitdeploy'] for x in batches[1::2]])

    server_batch_mapping = generate_server_batch_mapping(batches)
    for s in server_batch_mapping:
        if 'Consumer' in server_batch_mapping[s]:
            assert server_batch_mapping[s]['Configs'] < server_batch_mapping[s]['Consumer']
        elif 'ServiceBusLow' in server_batch_mapping[s]:
            assert server_batch_mapping[s]['Configs'] < server_batch_mapping[s]['ServiceBusLow']

def test_simulated_scorebig_unordered_groups():
    '''
    Test that unordered groups similar to ScoreBig setup are split into batches respecting gitdeploy ordering
    '''
    batches = elita.deployment.deploy.BatchCompute.compute_rolling_batches(divisor, None,
                                                                           simulated_scorebig_nonrolling_groups)

    assert len(batches) == 2
    assert all(["servers" in x and "gitdeploys" in x for x in batches])
    assert batches[0]['ordered_gitdeploy'] and not batches[1]['ordered_gitdeploy']

    server_batch_mapping = generate_server_batch_mapping(batches)
    for s in server_batch_mapping:
        if 'ServiceBusHigh' in server_batch_mapping[s]:
            assert server_batch_mapping[s]['Configs'] < server_batch_mapping[s]['ServiceBusHigh']
        elif 'ServiceBusLow' in server_batch_mapping[s]:
            assert server_batch_mapping[s]['Configs'] < server_batch_mapping[s]['ServiceBusLow']



if __name__ == '__main__':
    test_rolling_batches()
    test_rolling_and_nonrolling_batches()
    test_ordered_rolling_batches()
    test_ordered_and_unordered_rolling_batches()


import mock
#from elita.deployment import deploy, salt_control
import elita.deployment.deploy
import elita.deployment.salt_control
from elita.deployment import gitservice
from elita.models import DataService, BuildDataService, GitDataService, JobDataService, GroupDataService, \
    DeploymentDataService, ActionService
from elita.actions.action import RegisterHooks
#import logging


def return_gitdeploy(app, name):
    return {
        "application": app,
        "name": name,
        "package": "master",
        "location": {
            "gitrepo": {
                "name": "gr0"
            },
            "path": "/foo/bar",
            "default_branch": "master"
        },
        "servers": ["server0", "server1"]
    }


def return_build(app, name):
    return {
        "app_name": app,
        "build_name": name,
        "packages": {
            "master": {
                "file_type": "zip",
                "filename": "/foo/bar/baz.zip"
            }
        }
    }


def setup_mock_datasvc():
    mock_datasvc = mock.Mock(spec=DataService)
    mock_datasvc.attach_mock(mock.Mock(spec=BuildDataService), "buildsvc")
    mock_datasvc.attach_mock(mock.Mock(spec=GitDataService), "gitsvc")
    mock_datasvc.attach_mock(mock.Mock(spec=JobDataService), "jobsvc")
    mock_datasvc.attach_mock(mock.Mock(spec=DeploymentDataService), "deploysvc")
    mock_datasvc.attach_mock(mock.Mock(spec=ActionService), "actionsvc")
    mock_datasvc.actionsvc.attach_mock(mock.Mock(spec=RegisterHooks), "hooks")
    mock_datasvc.job_id = "fake_job_id"
    mock_datasvc.gitsvc.GetGitDeploy = return_gitdeploy
    mock_datasvc.buildsvc.GetBuild = return_build
    mock_datasvc.settings = {
        'elita.mongo.host': 'localhost',
        'elita.mongo.port': 0,
        'elita.mongo.db': 'none'
    }
    return mock_datasvc


@mock.patch('elita.deployment.salt_control.SaltController')
@mock.patch('elita.deployment.salt_control.RemoteCommands')
@mock.patch('elita.deployment.gitservice.GitDeployManager')
@mock.patch('elita.deployment.deploy.regen_datasvc')
def test_simple_deployment(mockRD, mockGitDeployManager, mockRemoteCommands, mockSaltController):
    '''
    Test simple deployment with two gitdeploys and two servers
    '''

    servers = ["server0", "server1"]
    gitdeploys = ["gd0", "gd1"]

    mock_datasvc = setup_mock_datasvc()
    mockRD.return_value = None, mock_datasvc
    mockGitDeployManager.last_build = "nobuild"

    dc = elita.deployment.deploy.DeployController(mock_datasvc, 'mock_id')

    ok, results = dc.run("example_app", "example_build", servers, gitdeploys, parallel=False)

    #TODO: implement real checks that mock objs are called how we expect


def return_group(app, name):
    return {
        "rolling_deploy": name == 'gp0',
        "gitdeploys": ["gd0", "gd1"] if name == 'gp0' else ["gd2", "gd3"],
        "servers": ["server0", "server1"] if name == 'gp0' else ["server2", "server3"]
    }


def return_group_servers(app, name):
    return ['rolling_server{}'.format(x) for x in range(0, 5)] if name == 'gp0' else ['nonrolling_server{}'.format(x) for x in range(0, 5)]

@mock.patch('elita.deployment.salt_control.SaltController')
@mock.patch('elita.deployment.salt_control.RemoteCommands')
@mock.patch('elita.deployment.gitservice.GitDeployManager')
@mock.patch('elita.deployment.deploy.regen_datasvc')
def test_rolling_deployment(mockRD, mockGitDeployManager, mockRemoteCommands, mockSaltController):
    '''
    Test rolling deployment with one rolling group and one non-rolling
    '''

    groups = ["group0", "group1"]
    environments = ["testing"]

    mock_datasvc = setup_mock_datasvc()
    mock_datasvc.groupsvc = mock.Mock(spec=GroupDataService)
    mock_datasvc.groupsvc.GetGroup = return_group
    mock_datasvc.groupsvc.GetGroupServers = return_group_servers
    mockRD.return_value = None, mock_datasvc

    dc = elita.deployment.deploy.DeployController(mock_datasvc, 'mock_id')
    rdc = elita.deployment.deploy.RollingDeployController(mock_datasvc, dc, 'mock_id')

    gp0 = return_group("example_app", "gp0")
    gp1 = return_group("example_app", "gp1")

    rdc.run("example_app", "example_build",
            {
                "groups": ["gp0", "gp1"],
                "environments": ["testing"],
                "gitdeploys": list(set(gp0['gitdeploys'] + gp1['gitdeploys'])),
                "servers": list(set(gp0['servers'] + gp1['servers']))
            }, 2, 0, 0, parallel=False)

    #assert False

if __name__ == '__main__':
    test_simple_deployment()
    test_rolling_deployment()



import collections
import hashlib
import uuid
import base64
import os.path
import sys
import shutil
import bson
import pprint
import datetime
import pytz
import logging
import copy

import elita
from elita.crypto import keypair
from elita.deployment import deploy, salt_control
import util
import util.type_check
from deployment.gitservice import EMBEDDED_YAML_DOT_REPLACEMENT, GitDeployManager
import elita_exceptions
from elita.actions.action import ActionService

# URL model:
# root/app/
# root/app/{app_name}/builds/
# root/app/{app_name}/builds/{build_name}
# root/app/{app_name}/environments/
# root/app/{app_name}/environments/{env_name}/deployments
# root/app/{app_name}/environments/{env_name}/deployments/{deployment_id}
# root/app/{app_name}/environments/{env_name}/servers
# root/app/{app_name}/environemnt/{env_name}/servers/{server_name}

#dummy class to pass to RootService
class Request:
    db = None

class MongoService:
    # logspam
    #__metaclass__ = elita.util.LoggingMetaClass

    def __init__(self, db):
        '''
        @type db = pymongo.database.Database
        '''
        assert db
        self.db = db

    def create_new(self, collection, keys, classname, doc, remove_existing=True):
        '''
        Creates new document in collection. Optionally, remove any existing according to keys (which specify how the
        new document is unique)

        Returns id of new document
        '''
        assert elita.util.type_check.is_string(collection)
        assert elita.util.type_check.is_dictlike(keys)
        assert elita.util.type_check.is_optional_str(classname)
        assert elita.util.type_check.is_dictlike(doc)
        assert collection
        # keys/classname are only mandatory if remove_existing=True
        assert (keys and classname and remove_existing) or not remove_existing
        if classname:
            doc['_class'] = classname
        existing = None
        if remove_existing:
            existing = [d for d in self.db[collection].find(keys)]
            for k in keys:
                doc[k] = keys[k]
            if '_id' in doc:
                del doc['_id']
        id = self.db[collection].save(doc, fsync=True)
        logging.debug("new id: {}".format(id))
        if existing and remove_existing:
            logging.warning("create_new found existing docs! deleting...(collection: {}, keys: {})".format(collection, keys))
            keys['_id'] = {'$ne': id}
            self.db[collection].remove(keys)
        return id

    def modify(self, collection, keys, path, doc_or_obj):
        '''
        Modifies document with the keys in doc. Does so atomically but remember that any key will overwrite the existing
        key.

        doc_or_obj could be None, zero, etc.

        Returns boolean indicating success
        '''
        assert hasattr(path, '__iter__')
        assert path
        assert elita.util.type_check.is_string(collection)
        assert isinstance(keys, dict)
        assert collection and keys
        dlist = [d for d in self.db[collection].find(keys)]
        assert dlist
        canonical_id = dlist[0]['_id']
        if len(dlist) > 1:
            logging.warning("Found duplicate entries for query {} in collection {}; using the first and removing others"
                            .format(keys, collection))
            keys['_id'] = {'$ne': canonical_id}
            self.db[collection].remove(keys)
        path_dot_notation = '.'.join(path)
        result = self.db[collection].update({'_id': canonical_id}, {'$set': {path_dot_notation: doc_or_obj}}, fsync=True)
        return result['n'] == 1 and result['updatedExisting'] and not result['err']

    def delete(self, collection, keys):
        '''
        Drop a document from the collection

        Return whatever pymongo returns for deletion
        '''
        assert elita.util.type_check.is_string(collection)
        assert isinstance(keys, dict)
        assert collection and keys
        dlist = [d for d in self.db[collection].find(keys)]
        assert dlist
        if len(dlist) > 1:
            logging.warning("Found duplicate entries for query {} in collection {}; removing all".format(keys,
                                                                                                        collection))
        return self.db[collection].remove(keys, fsync=True)

    def update_roottree(self, path, collection, id, doc=None):
        '''
        Update the root tree at path [must be a tuple of indices: ('app', 'myapp', 'builds', '123-foo')] with DBRef
        Optional doc can be passed in which will be inserted into the tree after adding DBRef field

        Return boolean indicating success
        '''
        assert hasattr(path, '__iter__')
        assert elita.util.type_check.is_string(collection)
        assert id.__class__.__name__ == 'ObjectId'
        assert util.type_check.is_optional_dict(doc)
        path_dot_notation = '.'.join(path)
        root_tree_doc = doc if doc else {}
        root_tree_doc['_doc'] = bson.DBRef(collection, id)
        result = self.db['root_tree'].update({}, {'$set': {path_dot_notation: root_tree_doc}}, fsync=True)
        return result['n'] == 1 and result['updatedExisting'] and not result['err']

    def rm_roottree(self, path):
        '''
        Delete/remove the root_tree reference at path
        '''
        assert hasattr(path, '__iter__')
        assert path
        path_dot_notation = '.'.join(path)
        result = self.db['root_tree'].update({}, {'$unset': {path_dot_notation: ''}}, fsync=True)
        return result['n'] == 1 and result['updatedExisting'] and not result['err']

    def get(self, collection, keys, multi=False, empty=False):
        '''
        Thin wrapper around find()
        Retrieve a document from Mongo, keyed by name. Optionally, if duplicates are found, delete all but the first.
        If empty, it's ok to return None if nothing matches

        Returns document
        @rtype: dict | list(dict) | None
        '''
        assert elita.util.type_check.is_string(collection)
        assert isinstance(keys, dict)
        assert collection
        dlist = [d for d in self.db[collection].find(keys)]
        assert dlist or empty
        if len(dlist) > 1 and not multi:
            logging.warning("Found duplicate entries ({}) for query {} in collection {}; dropping all but the first"
                            .format(len(dlist), keys, collection))
            keys['_id'] = {'$ne': dlist[0]['_id']}
            self.db[collection].remove(keys)
        return dlist if multi else (dlist[0] if dlist else dlist)

    def dereference(self, dbref):
        '''
        Simple wrapper around db.dereference()
        Returns document pointed to by DBRef

        @type id: bson.DBRef
        '''
        assert dbref
        assert dbref.__class__.__name__ == 'DBRef'
        return self.db.dereference(dbref)


class GenericChildDataService:
    __metaclass__ = elita.util.LoggingMetaClass

    def __init__(self, mongo_service, root, settings, job_id=None):
        '''
        @type mongo_service: MongoService
        @type root: RootTree
        @type settings: pyramid.registry.Registry
        @type job_id: None | str
        '''
        assert isinstance(mongo_service, MongoService)
        assert isinstance(root, RootTree)
        assert elita.util.type_check.is_optional_str(job_id)
        self.mongo_service = mongo_service
        self.root = root
        self.settings = settings
        self.job_id = job_id

    def populate_dependencies(self, dependency_objs):
        '''
        Child dataservice classes may need to access methods of siblings. This allows parent dataservice to inject
        cross dependencies as needed without generating a big reference cycle.

        dependency_objs = { 'FooDataService': FooDataService }

        @type dependency_objs: dict
        '''
        assert util.type_check.is_dictlike(dependency_objs)
        self.deps = dependency_objs

    def NewContainer(self, class_name, name, parent):
        '''
        Create new container object suitable for a root_tree reference
        '''
        assert class_name and name and parent
        assert elita.util.type_check.is_string(class_name)
        assert elita.util.type_check.is_string(name)
        assert elita.util.type_check.is_string(parent)
        return self.mongo_service.create_new('containers', {'name': name, 'parent': parent}, class_name,
                                             {'name': name, 'parent': parent}, remove_existing=False)

    def UpdateObject(self, collection, keys, doc):
        '''
        Generic method to update a particular object (document) with the data in doc
        '''
        assert collection and keys and doc
        assert elita.util.type_check.is_string(collection)
        assert elita.util.type_check.is_dictlike(keys)
        assert elita.util.type_check.is_dictlike(doc)

        paths = elita.util.paths_from_nested_dict(doc)
        assert paths
        for path in paths:
            self.mongo_service.modify(collection, keys, path[:-1], path[-1])

    def AddThreadLocalRootTree(self, path):
        '''
        Add new node to thread-local root_tree (copy from mongo's root_tree)
        '''
        assert path
        assert elita.util.type_check.is_seq(path)
        root_tree = self.mongo_service.get('root_tree', {})
        assert root_tree
        reduce(lambda d, k: d[k], path[:-1], self.root)[path[-1]] = reduce(lambda d, k: d[k], path, root_tree)

    def RmThreadLocalRootTree(self, path):
        '''
        Remove deleted node from thread-local root_tree
        '''
        assert path
        assert elita.util.type_check.is_seq(path)
        node = reduce(lambda d, k: d[k], path[:-1], self.root)
        del node[path[-1]]


class BuildDataService(GenericChildDataService):

    def GetBuilds(self, app_name):
        '''
        Get all builds for application.

        When getting a list of all objects of a given type, the convention is to pull from in-memory root_tree instead
        of directly from mongo. This keeps it fast so we can do it frequently for things like parameter validation, etc
        '''
        assert app_name
        assert elita.util.type_check.is_string(app_name)
        assert app_name in self.root['app']
        return [build for build in self.root['app'][app_name]['builds'] if build[0] != '_']

    def NewBuild(self, app_name, build_name, attribs):
        '''
        Create new build document
        '''
        assert elita.util.type_check.is_string(app_name)
        assert elita.util.type_check.is_string(build_name)
        assert isinstance(attribs, dict)
        assert app_name and build_name

        app_doc = self.mongo_service.get('applications', {'app_name': app_name})
        assert app_doc

        buildobj = Build({
            'app_name': app_name,
            'build_name': build_name,
            'attributes': attribs
        })

        bid = self.mongo_service.create_new('builds', {'app_name': app_name, 'build_name': build_name}, 'Build',
                                            buildobj.get_doc())
        self.mongo_service.update_roottree(('app', app_name, 'builds', build_name), 'builds', bid)
        self.AddThreadLocalRootTree(('app', app_name, 'builds', build_name))
        return True

    def AddPackages(self, app, build, packages):
        '''
        Add new packages fields to existing build. Regenerate legacy 'files' field (which is a flat array of files
        associated with build
        '''

        assert elita.util.type_check.is_string(app)
        assert elita.util.type_check.is_string(build)
        assert isinstance(packages, dict)
        assert app in self.root['app']
        assert build in self.root['app'][app]['builds']

        keys = {'app_name': app, 'build_name': build}
        for p in packages:
            path = ('packages', p)
            self.mongo_service.modify('builds', keys, path, packages[p])

        build_doc = self.mongo_service.get('builds', keys)
        assert all([p in build_doc['packages'] for p in packages])

        #generate files from packages (avoid dupes)
        files = [{"file_type": packages[p]['file_type'], "path": packages[p]['filename']} for p in packages]
        self.mongo_service.modify('builds', keys, ('files',), files)

    def UpdateBuild(self, app, name, doc):
        '''
        Update build with keys in doc.
        '''
        assert elita.util.type_check.is_string(app)
        assert elita.util.type_check.is_string(name)
        assert isinstance(doc, dict)
        assert app and name
        assert app in self.root['app']
        assert name in self.root['app'][app]['builds']

        self.UpdateObject('builds', {'app_name': app, 'build_name': name}, doc)

    def DeleteBuildStorage(self, app_name, build_name):
        '''
        Delete all stored files associated with build.
        '''
        assert elita.util.type_check.is_string(app_name)
        assert elita.util.type_check.is_string(build_name)
        assert app_name and build_name

        app_doc = self.mongo_service.get('applications', {'app_name': app_name})
        assert app_doc

        dir = self.settings['elita.builds.dir']
        path = "{root_dir}/{app}/{build}".format(root_dir=dir, app=app_name, build=build_name)
        logging.debug("DeleteBuildStorage: path: {}".format(path))
        if os.path.isdir(path):
            logging.debug("DeleteBuildStorage: remove_build: deleting")
            shutil.rmtree(path)

    def DeleteBuild(self, app_name, build_name):
        '''
        Delete build object, root_tree reference and all stored files.
        '''
        assert elita.util.type_check.is_string(app_name)
        assert elita.util.type_check.is_string(build_name)
        assert app_name and build_name

        root_path = ('app', app_name, 'builds', build_name)
        self.mongo_service.rm_roottree(root_path)
        self.RmThreadLocalRootTree(root_path)
        self.mongo_service.delete('builds', {'app_name': app_name, 'build_name': build_name})
        self.DeleteBuildStorage(app_name, build_name)

    def GetBuild(self, app_name, build_name):
        '''
        Get build document
        '''
        assert elita.util.type_check.is_string(app_name)
        assert elita.util.type_check.is_string(build_name)
        assert app_name and build_name
        assert app_name in self.root['app']
        assert build_name in self.root['app'][app_name]['builds']
        doc = self.mongo_service.get('builds', {'app_name': app_name, 'build_name': build_name})
        doc['created_datetime'] = doc['_id'].generation_time
        return {k: doc[k] for k in doc if k[0] != '_'}


class UserDataService(GenericChildDataService):

    def NewUser(self, name, pw, perms, attribs):
        '''
        Create a new user object and insert root_tree references for both the user and the computed permissions
        endpoint. Pipe parameters into User object to get the pw hashed, etc.
        '''
        assert elita.util.type_check.is_string(name)
        assert elita.util.type_check.is_string(pw)
        assert isinstance(attribs, dict)
        assert isinstance(perms, dict)
        assert name and pw and perms

        userobj = User({
            'username': name,
            'permissions': perms,
            'password': pw,
            'attributes': attribs
        })
        uid = self.mongo_service.create_new('users', {'username': userobj.username}, 'User', userobj.get_doc())
        pid = self.mongo_service.create_new('userpermissions', {'username': userobj.username}, 'UserPermissions', {
            "username": userobj.username,
            "applications": list(),
            "actions": dict(),
            "servers": list()
        })
        self.mongo_service.update_roottree(('global', 'users', userobj.username), 'users', uid)
        self.AddThreadLocalRootTree(('global', 'users', userobj.username))
        self.mongo_service.update_roottree(('global', 'users', userobj.username, 'permissions'), 'userpermissions', pid)
        self.AddThreadLocalRootTree(('global', 'users', userobj.username, 'permissions'))
        return uid, pid

    def GetUserTokens(self, username):
        '''
        Get all auth tokens associated with user
        '''
        assert elita.util.type_check.is_string(username)
        assert username
        return [d['token'] for d in self.mongo_service.get('tokens', {'username': username}, multi=True, empty=True)]

    def GetUserFromToken(self, token):
        '''
        Get username associated with token
        '''
        assert elita.util.type_check.is_string(token)
        assert token
        return self.mongo_service.get('tokens', {'token': token})['username']

    def GetAllTokens(self):
        '''
        Get all valid tokens
        '''
        return [token for token in self.root['global']['tokens'] if token[0] != '_']

    def NewToken(self, username):
        '''
        Create new auth token associated with username and insert reference into root_tree
        '''
        assert elita.util.type_check.is_string(username)
        assert username

        token = Token({
            'username': username
        })
        tid = self.mongo_service.create_new('tokens', {'username': username, 'token': token.token}, 'Token',
                                            token.get_doc())
        self.mongo_service.update_roottree(('global', 'tokens', token.token), 'tokens', tid)
        self.AddThreadLocalRootTree(('global', 'tokens', token.token))
        return token

    def GetUsers(self):
        '''
        Get all valid users
        '''
        return [user for user in self.root['global']['users'] if user[0] != '_']

    def GetUser(self, username):
        '''
        Get user document
        '''
        assert elita.util.type_check.is_string(username)
        assert username
        doc = self.mongo_service.get('users', {'username': username})
        doc['created_datetime'] = doc['_id'].generation_time
        return {k: doc[k] for k in doc if k[0] != '_'}

    def DeleteUser(self, name):
        '''
        Delete a single user and root_tree reference
        '''
        assert name
        assert elita.util.type_check.is_string(name)
        self.mongo_service.rm_roottree(('global', 'users', name))
        self.RmThreadLocalRootTree(('global', 'users', name))
        self.mongo_service.delete('users', {'username': name})

    def DeleteToken(self, token):
        '''
        Delete a token and root_tree reference
        '''
        assert token
        assert elita.util.type_check.is_string(token)
        self.mongo_service.rm_roottree(('global', 'tokens', token))
        self.RmThreadLocalRootTree(('global', 'tokens', token))
        self.mongo_service.delete('tokens', {'token': token})

    def UpdateUser(self, name, doc):
        '''
        Update user with keys in doc
        '''
        assert name and doc
        assert elita.util.type_check.is_string(name)
        assert elita.util.type_check.is_dictlike(doc)
        assert name in self.root['global']['users']

        if "password" in doc:
            user = User(doc)
            doc['hashed_pw'] = user.hashed_pw
            doc['salt'] = user.salt
            doc['password'] = None

        self.UpdateObject('users', {'username': name}, doc)


class ApplicationDataService(GenericChildDataService):
    def GetApplications(self):
        '''
        Get all applications. Pull from in-memory root_tree rather than mongo for speed.
        '''
        return [app for app in self.root['app'] if app[0] != '_']

    def GetApplication(self, app_name):
        '''
        Get application document
        '''
        assert app_name
        assert elita.util.type_check.is_string(app_name)
        doc = self.mongo_service.get('applications', {'app_name': app_name})
        doc['created_datetime'] = doc['_id'].generation_time
        return {k: doc[k] for k in doc if k[0] != '_'}

    def NewApplication(self, app_name):
        '''
        Create new application and all subcontainers and root_tree sub-references
        '''
        assert app_name
        assert elita.util.type_check.is_string(app_name)
        aid = self.mongo_service.create_new('applications', {'app_name': app_name}, 'Application', {})
        root_doc = {
            "builds": {"_doc": bson.DBRef('containers', self.NewContainer("BuildContainer", "builds", app_name))},
            "actions": {"_doc": bson.DBRef('containers', self.NewContainer("ActionContainer", "actions", app_name))},
            "gitrepos": {"_doc": bson.DBRef('containers', self.NewContainer("GitRepoContainer", "gitrepos", app_name))},
            "gitdeploys": {"_doc": bson.DBRef('containers', self.NewContainer("GitDeployContainer", "gitdeploys", app_name))},
            "deployments": {"_doc": bson.DBRef('containers', self.NewContainer("DeploymentContainer", "deployments", app_name))},
            "groups": {"_doc": bson.DBRef('containers', self.NewContainer("GroupContainer", "groups", app_name))},
            "packagemaps": {"_doc": bson.DBRef('containers', self.NewContainer("PackageMapContainer", "packagemaps", app_name))}
        }
        res = self.mongo_service.update_roottree(('app', app_name), 'applications', aid, doc=root_doc)
        self.AddThreadLocalRootTree(('app', app_name))
        return res

    def DeleteApplication(self, app_name):
        '''
        Delete application and all root_tree references and sub-objects.
        '''
        assert app_name
        assert elita.util.type_check.is_string(app_name)
        self.mongo_service.rm_roottree(('app', app_name))
        self.mongo_service.delete('applications', {'app_name': app_name})
        self.mongo_service.delete('builds', {'app_name': app_name})
        self.mongo_service.delete('gitrepos', {'application': app_name})
        self.mongo_service.delete('gitdeploys', {'application': app_name})
        self.mongo_service.delete('deployments', {'application': app_name})
        self.mongo_service.delete('groups', {'application': app_name})
        self.mongo_service.delete('packagemaps', {'application': app_name})
        self.RmThreadLocalRootTree(('app', app_name))

    def GetApplicationCensus(self, app_name):
        '''
        Generates a census of all environments, groups, servers and builds deployed:
        {
          "env_name": {
            "group_name": {
                "server_name": {
                    "gitdeploy_name": {
                        "committed": "build_name",
                        "deployed": "build_name"
                }
            }
        }
        '''
        assert app_name
        assert elita.util.type_check.is_string(app_name)
        groups = [d['name'] for d in self.mongo_service.get('groups', {'application': app_name}, multi=True, empty=True)]
        envs = list({d['environment'] for d in self.mongo_service.get('servers', {}, multi=True, empty=True)})
        census = dict()
        for e in envs:
            census[e] = dict()
            for g in groups:
                g_servers = self.deps['GroupDataService'].GetGroupServers(app_name, g, environments=[e])
                census[e][g] = dict()
                for s in g_servers:
                    census[e][g][s] = dict()
                    group_doc = self.deps['GroupDataService'].GetGroup(app_name, g)
                    for gd in group_doc['gitdeploys']:
                        gdl = gd if isinstance(gd, list) else [gd]
                        for gd in gdl:
                            gd_doc = self.deps['GitDataService'].GetGitDeploy(app_name, gd)
                            census[e][g][s][gd] = {
                                "committed": gd_doc['location']['gitrepo']['last_build'],
                                "deployed": gd_doc['deployed_build']
                            }
                    if len(census[e][g][s]) == 0:
                        del census[e][g][s]
                if len(census[e][g]) == 0:
                    del census[e][g]
            if len(census[e]) == 0:
                del census[e]
        return census

    def UpdateApplication(self, app, doc):
        '''
        Update application with keys in doc
        '''
        assert app and doc
        assert elita.util.type_check.is_string(app)
        assert elita.util.type_check.is_dictlike(doc)
        assert app in self.root['app']

        self.UpdateObject('applications', {'app_name': app}, doc)

class PackageMapDataService(GenericChildDataService):
    def GetPackageMaps(self, app):
        '''
        Get all packagemaps for application
        '''
        assert app
        assert elita.util.type_check.is_string(app)
        assert app in self.root['app']
        return [pm for pm in self.root['app'][app]['packagemaps'] if pm[0] != '_']

    def GetPackageMap(self, app, name):
        '''
        Get document for packagemap
        '''
        assert app and name
        assert elita.util.type_check.is_string(app)
        assert elita.util.type_check.is_string(name)
        doc = self.mongo_service.get('packagemaps', {'application': app, 'name': name})
        doc['created_datetime'] = doc['_id'].generation_time
        return {k: doc[k] for k in doc if k[0] != '_'}

    def NewPackageMap(self, app, name, packages, attributes=None):
        '''
        Create new packagemap

        @type app: str
        @type name: str
        @type packages: dict
        @type attributes: dict | None
        '''
        assert app and name and packages
        assert elita.util.type_check.is_string(app)
        assert elita.util.type_check.is_string(name)
        assert elita.util.type_check.is_dictlike(packages)
        assert elita.util.type_check.is_optional_dict(attributes)
        attributes = attributes if attributes else {}
        pm = PackageMap({
            'application': app,
            'name': name,
            'packages': packages,
            'attributes': attributes
        })
        pmid = self.mongo_service.create_new('packagemaps', {'application': app, 'name': name}, 'PackageMap', pm.get_doc())
        self.mongo_service.update_roottree(('app', app, 'packagemaps', name), 'packagemaps', pmid)
        self.AddThreadLocalRootTree(('app', app, 'packagemaps', name))

    def DeletePackageMap(self, app, name):
        '''
        Delete a packagemap object and root_tree reference
        '''
        assert app and name
        assert elita.util.type_check.is_string(app)
        assert elita.util.type_check.is_string(name)
        self.mongo_service.rm_roottree(('app', app, 'packagemaps', name))
        self.RmThreadLocalRootTree(('app', app, 'packagemaps', name))
        self.mongo_service.delete('packagemaps', {'application': app, 'name': name})

    def UpdatePackageMap(self, app, name, doc):
        '''
        Update packagemap with keys in doc
        '''
        assert app and name and doc
        assert elita.util.type_check.is_string(app)
        assert elita.util.type_check.is_string(name)
        assert elita.util.type_check.is_dictlike(doc)
        assert app in self.root['app']
        assert name in self.root['app'][app]['packagemaps']

        self.UpdateObject('packagemaps', {'application': app, 'name': name}, doc)

class GroupDataService(GenericChildDataService):
    def GetGroups(self, app):
        '''
        Get all groups for application.
        '''
        assert app
        assert elita.util.type_check.is_string(app)
        assert app in self.root['app']
        return [group for group in self.root['app'][app]['groups'] if group[0] != '_']

    def GetGroup(self, app, name):
        '''
        Get document for application group
        '''
        assert app and name
        assert elita.util.type_check.is_string(app)
        assert elita.util.type_check.is_string(name)
        doc = self.mongo_service.get('groups', {'application': app, 'name': name})
        doc['servers'] = self.GetGroupServers(app, name, group_doc=doc)
        doc['created_datetime'] = doc['_id'].generation_time
        return {k: doc[k] for k in doc if k[0] != '_'}

    def NewGroup(self, app, name, gitdeploys, rolling_deploy=False, description="", attributes=None):
        '''
        Create new application group.

        @type app: str
        @type name: str
        @type gitdeploys: list[str]
        @type rolling_deploys: True | False
        @type description: str
        @type attributes: dict | None
        '''
        assert app and name and gitdeploys
        assert elita.util.type_check.is_string(app)
        assert elita.util.type_check.is_string(name)
        assert elita.util.type_check.is_seq(gitdeploys)
        assert elita.util.type_check.is_string(description)
        assert elita.util.type_check.is_optional_dict(attributes)
        attributes = attributes if attributes else {}
        gp = Group({
            "application": app,
            "name": name,
            "description": description,
            "gitdeploys": gitdeploys,
            "attributes": attributes,
            "rolling_deploy": rolling_deploy
        })

        gid = self.mongo_service.create_new('groups', {'application': app, 'name': name}, 'Group', gp.get_doc())
        self.mongo_service.update_roottree(('app', app, 'groups', name), 'groups', gid)
        self.AddThreadLocalRootTree(('app', app, 'groups', name))

    def DeleteGroup(self, app, name):
        '''
        Delete a group object and root_tree reference
        '''
        assert app and name
        assert elita.util.type_check.is_string(app)
        assert elita.util.type_check.is_string(name)
        self.mongo_service.rm_roottree(('app', app, 'groups', name))
        self.RmThreadLocalRootTree(('app', app, 'groups', name))
        self.mongo_service.delete('groups', {'application': app, 'name': name})

    def GetGroupServers(self, app, name, environments=None, group_doc=None):
        '''
        Build sets from initialized servers in each gitdeploy in the group, then take intersection of all the sets
        If environments specified, take intersection with that set as well

        Allow caller to provide group_doc to prevent infinite recursion from GetGroup calling GetGroupServers
        '''
        assert app and name
        assert elita.util.type_check.is_string(app)
        assert elita.util.type_check.is_string(name)
        assert app in self.root['app']
        assert name in self.root['app'][app]['groups']
        if not group_doc:
            group = self.GetGroup(app, name)
        else:
            group = group_doc
        # this is ugly. gitdeploys can either be a list of strings or a list of lists. We have to flatten the
        # list of lists if necessary
        server_sets = [set(self.deps['GitDataService'].GetGitDeploy(app, gd)['servers']) for sublist in group['gitdeploys'] for gd in sublist] if isinstance(group['gitdeploys'][0], list) else [set(self.deps['GitDataService'].GetGitDeploy(app, gd)['servers']) for gd in group['gitdeploys']]
        if environments:
            server_env_set = set()
            envs = self.deps['ServerDataService'].GetEnvironments()
            for e in environments:
                assert e in envs
                server_env_set = set(envs[e]).union(server_env_set)
            server_sets.append(server_env_set)
        return list(set.intersection(*server_sets))

    def UpdateGroup(self, app, name, doc):
        '''
        Update group with keys in doc
        '''
        assert app and name and doc
        assert elita.util.type_check.is_string(app)
        assert elita.util.type_check.is_string(name)
        assert elita.util.type_check.is_dictlike(doc)
        assert app in self.root['app']
        assert name in self.root['app'][app]['groups']

        self.UpdateObject('groups', {'application': app, 'name': name}, doc)

class JobDataService(GenericChildDataService):
    def GetAllActions(self, app_name):
        '''
        Get all actions associated with application. Get it from root_tree because the actions are dynamically populated
        at the start of each request.
        '''
        assert app_name
        assert elita.util.type_check.is_string(app_name)
        if 'actions' in self.root['app'][app_name]:
            return [action for action in self.root['app'][app_name]['actions'] if action[0] != '_']

    def GetAction(self, app_name, action_name):
        '''
        Get details (name, description, parameters) about all actions associated with application
        '''
        assert app_name and action_name
        assert elita.util.type_check.is_string(app_name)
        assert elita.util.type_check.is_string(action_name)
        actions = self.deps['ActionService'].get_action_details(app_name, action_name)
        return {k: actions[k] for k in actions if k is not "callable"}

    def NewJob(self, name, job_type, data):
        '''
        Create new job object.

        data parameter can be anything serializable (including None) so that's all we will check for
        '''
        assert name and job_type
        assert elita.util.type_check.is_serializable(data)
        job = Job({
            'status': "running",
            'name': name,
            'job_type': job_type,
            'data': data,
            'attributes': {
                'name': name
            }
        })
        jid = self.mongo_service.create_new('jobs', {'job_id': str(job.job_id)}, 'Job', job.get_doc())
        self.mongo_service.update_roottree(('job', str(job.job_id)), 'jobs', jid)
        self.AddThreadLocalRootTree(('job', str(job.job_id)))
        return job

    def NewJobData(self, data):
        '''
        Insert new job_data record. Called by async jobs to log progress. Data inserted here can be viewed by user
        by polling the respective job object endpoint.

        Only valid to be called in an asynch context, so assert we have a valid job_id
        '''
        assert data
        assert elita.util.type_check.is_serializable(data)
        assert self.job_id
        elita.util.change_dict_keys(data, '.', '_')
        self.mongo_service.create_new('job_data', {}, None, {
            'job_id': self.job_id,
            'data': data
        }, remove_existing=False)

    def GetJobs(self, active):
        '''
        Get all actively running jobs. Pulling from mongo could possibly be more efficient (maybe) than using
        in-memory root_tree because we're querying on the status field
        '''
        return [d['job_id'] for d in self.mongo_service.get('jobs', {'status': 'running'}, multi=True)]

    def GetJobData(self, job_id):
        '''
        Get job data for a specific job sorted by created_datetime (ascending)
        '''
        assert job_id
        assert elita.util.type_check.is_string(job_id)
        return sorted([{'created_datetime': d['_id'].generation_time.isoformat(' '), 'data': d['data']} for
                       d in self.mongo_service.get('job_data', {'job_id': job_id}, multi=True, empty=True)],
                      key=lambda k: k['created_datetime'])

    def SaveJobResults(self, results):
        '''
        Called at the end of async jobs. Changes state of job object to reflect job completion.
        '''
        assert self.job_id
        assert elita.util.type_check.is_serializable(results)
        now = datetime.datetime.now(tz=pytz.utc)
        doc = self.mongo_service.get('jobs', {'job_id': self.job_id})
        assert doc and elita.util.type_check.is_dictlike(doc) and '_id' in doc
        results_sanitized = copy.deepcopy(results)
        elita.util.change_dict_keys(results_sanitized, '.', '_')
        diff = (now - doc['_id'].generation_time).total_seconds()
        self.mongo_service.modify('jobs', {'job_id': self.job_id}, ('status',), "completed")
        self.mongo_service.modify('jobs', {'job_id': self.job_id}, ('completed_datetime',), now)
        self.mongo_service.modify('jobs', {'job_id': self.job_id}, ('duration_in_seconds',), diff)
        self.NewJobData({"completed_results": results_sanitized})

    def NewAction(self, app_name, action_name, params):
        '''
        Register new dynamically-loaded action in root_tree. These are loaded from plugins at the start of each request
        Note that action is added to in-memory root_tree object (not the root_tree record in mongo) because it is not
        persistent. Note further that this is *our* (meaning this thread's) root_tree and will only be in effect for the
        duration of this request, so we don't care about any root_tree updates by other threads running concurrently
        '''
        assert app_name and action_name
        assert elita.util.type_check.is_string(app_name)
        assert elita.util.type_check.is_string(action_name)
        assert elita.util.type_check.is_optional_seq(params)
        logging.debug("NewAction: app_name: {}".format(app_name))
        logging.debug("NewAction: action_name: {}".format(action_name))
        if app_name in self.deps['ApplicationDataService'].GetApplications():
            assert app_name in self.root['app']
            assert 'actions' in self.root['app'][app_name]
            assert elita.util.type_check.is_dictlike(self.root['app'][app_name]['actions'])
            self.root['app'][app_name]['actions'][action_name] = Action(app_name, action_name, params, self)
        else:
            logging.debug("NewAction: application '{}' not found".format(app_name))

    def ExecuteAction(self, app_name, action_name, params):
        '''
        Spawn async job for an action
        '''
        assert app_name and action_name and params
        assert elita.util.type_check.is_string(app_name)
        assert elita.util.type_check.is_string(action_name)
        assert elita.util.type_check.is_dictlike(params)
        return self.deps['ActionService'].async(app_name, action_name, params)

class ServerDataService(GenericChildDataService):
    def GetServers(self):
        '''
        Return a list of all extant server objects
        '''
        return [k for k in self.root['server'].keys() if k[0] != '_' and k != 'environments']

    def NewServer(self, name, attribs, environment, existing=False):
        '''
        Create a new server object
        '''
        server = Server({
            'name': name,
            'status': 'new',
            'server_type': 'unknown',
            'environment': environment,
            'attributes': attribs
        })
        sid = self.mongo_service.create_new('servers', {'name': name}, 'Server', server.get_doc())
        self.mongo_service.update_roottree(('server', name), 'servers', sid, doc={
            "gitdeploys": self.NewContainer('GitDeployContainer', name, "gitdeploys")
        })
        self.AddThreadLocalRootTree(('server', name))
        return {
            'NewServer': {
                'name': name,
                'environment': environment,
                'attributes': attribs,
                'status': 'ok'
            }
        }

    def UpdateServer(self, name, doc):
        '''
        Change existing server object with data in doc
        '''
        assert name and doc
        assert elita.util.type_check.is_string(name)
        assert elita.util.type_check.is_dictlike(doc)
        assert name in self.root['server']
        prototype = Server({})  # get all valid default top-level properties
        assert all([key in prototype.get_doc() for key in doc])

        self.UpdateObject('servers', {'name': name}, doc)

    def DeleteServer(self, name):
        '''
        Delete a server object
        '''
        assert name
        assert elita.util.type_check.is_string(name)
        assert name in self.root['server']
        self.mongo_service.rm_roottree(('server', name))
        self.RmThreadLocalRootTree(('server', name))
        self.mongo_service.delete('servers', {'name': name})

    def GetGitDeploys(self, name):
        '''
        Get all gitdeploys initialized on a server. The canonical data source is the gitdeploy object (which contains
        a list of servers it's been initialized on)
        '''
        assert name
        assert elita.util.type_check.is_string(name)
        assert name in self.root['server']
        gitdeploys = self.mongo_service.get('gitdeploys', {'servers': {'$in': [name]}}, multi=True, empty=True)
        if gitdeploys:
            return [{'application': gd['application'], 'gitdeploy_name': gd['name']} for gd in gitdeploys]
        else:
            return []

    def GetEnvironments(self):
        '''
        Get a census of all environments. "environment" is just a tag associated with a server upon creation, so get all
        tags and dedupe.
        '''
        environments = dict()
        for sd in self.mongo_service.get('servers', {}, multi=True):
            assert elita.util.type_check.is_dictlike(sd)
            assert 'environment' in sd
            if sd['environment'] not in environments:
                environments[sd['environment']] = list()
            environments[sd['environment']].append(sd['name'])
        return environments

class GitDataService(GenericChildDataService):
    def NewGitDeploy(self, name, app_name, package, options, actions, location, attributes):
        '''
        Create new gitdeploy object. One of a few New* methods that returns status of success/failure to the view layer
        @rtype dict
        '''
        assert name and app_name and location
        assert elita.util.type_check.is_string(name)
        assert elita.util.type_check.is_string(app_name)
        assert elita.util.type_check.is_dictlike(location)
        assert all([k in location for k in ('gitrepo', 'path', 'default_branch')])
        assert app_name in self.root['app']
        assert elita.util.type_check.is_optional_str(package)
        assert elita.util.type_check.is_optional_dict(options)
        assert elita.util.type_check.is_optional_dict(actions)
        assert elita.util.type_check.is_optional_dict(attributes)

        #get associated gitrepo
        gitrepo_doc = self.mongo_service.get('gitrepos', {'name': location['gitrepo']})
        logging.debug("NewGitDeploy: gitrepo_doc: {}".format(gitrepo_doc))
        if not gitrepo_doc:
            return {'error': "invalid gitrepo (not found)"}

        #replace gitrepo name with DBRef
        location['gitrepo'] = bson.DBRef("gitrepos", gitrepo_doc['_id'])

        #construct gitdeploy document
        gd_obj = GitDeploy({})
        gd_doc = gd_obj.get_doc()
        gd_doc['name'] = name
        gd_doc['application'] = app_name
        gd_doc['location'] = location
        gd_doc['attributes'] = attributes if attributes else {}
        gd_doc['package'] = package

        #override defaults if specified
        if options:
            for k in options:
                if k in gd_doc['options']:
                    gd_doc['options'][k] = options[k]
        if actions:
            util.change_dict_keys(actions, '.', EMBEDDED_YAML_DOT_REPLACEMENT)
            for k in actions:
                if k in gd_doc['actions']:
                    gd_doc['actions'][k] = actions[k]
        gdid = self.mongo_service.create_new('gitdeploys', {'name': name, 'application': app_name}, 'GitDeploy', gd_doc)
        self.mongo_service.update_roottree(('app', app_name, 'gitdeploys', name), 'gitdeploys', gdid)
        self.AddThreadLocalRootTree(('app', app_name, 'gitdeploys', name))
        return {"ok": "done"}

    def GetGitDeploys(self, app):
        '''
        Get all gitdeploys associated with application

        @rtype: list
        '''
        assert app
        assert elita.util.type_check.is_string(app)
        assert app in self.root['app']
        return [k for k in self.root['app'][app]['gitdeploys'] if k[0] != '_']


    def GetGitDeploy(self, app, name):
        '''
        Get gitdeploy document. Dereference embedded DBrefs for convenience.

        @rtype: dict
        '''
        assert app and name
        assert elita.util.type_check.is_string(app)
        assert elita.util.type_check.is_string(name)
        assert app in self.root['app']
        assert name in self.root['app'][app]['gitdeploys']

        doc = self.mongo_service.get('gitdeploys', {'name': name, 'application': app})
        doc['created_datetime'] = doc['_id'].generation_time
        assert 'location' in doc
        assert 'gitrepo' in doc['location']
        #dereference embedded dbrefs
        doc['location']['gitrepo'] = self.mongo_service.dereference(doc['location']['gitrepo'])
        assert doc['location']['gitrepo']
        assert all([k in doc['location']['gitrepo'] for k in ('keypair', 'gitprovider')])
        doc['location']['gitrepo']['keypair'] = self.mongo_service.dereference(doc['location']['gitrepo']['keypair'])
        assert doc['location']['gitrepo']['keypair']
        doc['location']['gitrepo']['gitprovider'] = self.mongo_service.dereference(doc['location']['gitrepo']['gitprovider'])
        assert doc['location']['gitrepo']['gitprovider']
        return {k: doc[k] for k in doc if k[0] != '_'}

    def UpdateGitDeploy(self, app, name, doc):
        '''
        Update gitdeploy object with the data in doc
        '''
        assert app and name and doc
        assert elita.util.type_check.is_string(app)
        assert elita.util.type_check.is_string(name)
        assert elita.util.type_check.is_dictlike(doc)
        assert app in self.root['app']
        assert name in self.root['app'][app]['gitdeploys']

        #clean up any actions
        if 'actions' in doc:
            util.change_dict_keys(doc['actions'], '.', EMBEDDED_YAML_DOT_REPLACEMENT)

        #replace gitrepo with DBRef if necessary
        if 'location' in doc and 'gitrepo' in doc['location']:
            grd = self.mongo_service.get('gitrepos', {'name': doc['location']['gitrepo'], 'application': app})
            assert grd
            doc['location']['gitrepo'] = bson.DBRef('gitrepos', grd['_id'])

        self.UpdateObject('gitdeploys', {'name': name, 'application': app}, doc)

    def DeleteGitDeploy(self, app, name):
        '''
        Delete a gitdeploy object and the root_tree reference
        '''
        assert app and name
        assert elita.util.type_check.is_string(app)
        assert elita.util.type_check.is_string(name)
        assert app in self.root['app']
        assert name in self.root['app'][app]['gitdeploys']

        self.mongo_service.rm_roottree(('app', app, 'gitdeploys', name))
        self.RmThreadLocalRootTree(('app', app, 'gitdeploys', name))
        self.mongo_service.delete('gitdeploys', {'name': name, 'application': app})

    def GetGitProviders(self):
        '''
        Get all gitproviders

        @rtype: list(str) | None
        '''
        return [k for k in self.root['global']['gitproviders'] if k[0] != '_']

    def GetGitProvider(self, name):
        '''
        Get gitprovider document

        @rtype: dict | None
        '''
        assert name
        assert elita.util.type_check.is_string(name)
        assert name in self.root['global']['gitproviders']

        doc = self.mongo_service.get('gitproviders', {'name': name})
        doc['created_datetime'] = doc['_id'].generation_time
        return {k: doc[k] for k in doc if k[0] != '_'}

    def NewGitProvider(self, name, provider_type, auth):
        '''
        Create new gitprovider object
        '''
        assert name and type and auth
        assert elita.util.type_check.is_string(name)
        assert elita.util.type_check.is_string(provider_type)
        assert provider_type in ('bitbucket', 'github')
        assert elita.util.type_check.is_dictlike(auth)
        assert 'username' in auth and 'password' in auth
        assert elita.util.type_check.is_string(auth['username'])
        assert elita.util.type_check.is_string(auth['password'])

        gpo = GitProvider({
            'name': name,
            'type': provider_type,
            'auth': auth
        })

        gpid = self.mongo_service.create_new('gitproviders', {'name': name}, 'GitProvider', gpo.get_doc())
        self.mongo_service.update_roottree(('global', 'gitproviders', name), 'gitproviders', gpid)
        self.AddThreadLocalRootTree(('global', 'gitproviders', name))

    def UpdateGitProvider(self, name, doc):
        '''
        Modify gitprovider with the data in doc
        '''
        assert name and doc
        assert elita.util.type_check.is_string(name)
        assert elita.util.type_check.is_dictlike(doc)
        assert name in self.root['global']['gitproviders']

        self.UpdateObject('gitproviders', {'name': name}, doc)

    def DeleteGitProvider(self, name):
        '''
        Delete gitprovider object and root_tree reference
        '''
        assert name
        assert elita.util.type_check.is_string(name)
        assert name in self.root['global']['gitproviders']

        self.mongo_service.rm_roottree(('global', 'gitproviders', name))
        self.RmThreadLocalRootTree(('global', 'gitproviders', name))
        self.mongo_service.delete('gitproviders', {'name': name})

    def GetGitRepos(self, app):
        '''
        Get a list of all gitrepos associated with application

        @rtype: list(dict)
        '''
        assert app
        assert elita.util.type_check.is_string(app)
        assert app in self.root['app']

        return [k for k in self.root['app'][app]['gitrepos'] if k[0] != '_']

    def GetGitRepo(self, app, name):
        '''
        Get gitrepo document

        @rtype: dict
        '''
        assert app and name
        assert elita.util.type_check.is_string(app)
        assert elita.util.type_check.is_string(name)
        assert app in self.root['app']
        assert name in self.root['app'][app]['gitrepos']

        doc = self.mongo_service.get('gitrepos', {'name': name, 'application': app})
        assert doc and 'gitprovider' in doc and 'keypair' in doc
        doc['created_datetime'] = doc['_id'].generation_time
        gitprovider = self.mongo_service.dereference(doc['gitprovider'])
        doc['gitprovider'] = {k: gitprovider[k] for k in gitprovider if k[0] != '_'}
        keypair = self.mongo_service.dereference(doc['keypair'])
        doc['keypair'] = {k: keypair[k] for k in keypair if k[0] != '_'}
        return {k: doc[k] for k in doc if k[0] != '_'}

    def NewGitRepo(self, app, name, keypair, gitprovider, uri):
        '''
        Create new gitrepo object

        @rtype: dict
        '''
        assert app and name and keypair and gitprovider
        assert all([elita.util.type_check.is_string(p) for p in (app, name, keypair, gitprovider)])
        assert elita.util.type_check.is_optional_str(uri)
        assert app in self.root['app']
        assert keypair in self.root['global']['keypairs']
        assert gitprovider in self.root['global']['gitproviders']

        #get docs so we can generate DBRefs
        kp_doc = self.mongo_service.get('keypairs', {'name': keypair})
        gp_doc = self.mongo_service.get('gitproviders', {'name': gitprovider})
        if not gp_doc:
            return {'NewGitRepo': "gitprovider '{}' is unknown".format(gitprovider)}
        if not kp_doc:
            return {'NewGitRepo': "keypair '{}' is unknown".format(keypair)}

        gro = GitRepo({
            'name': name,
            'application': app,
            'keypair': bson.DBRef("keypairs", kp_doc['_id']),
            'gitprovider': bson.DBRef("gitproviders", gp_doc['_id']),
            'uri': uri,
            'last_build': None
        })

        grid = self.mongo_service.create_new('gitrepos', {'name': name, 'application': app}, 'GitRepo', gro.get_doc())
        self.mongo_service.update_roottree(('app', app, 'gitrepos', name), 'gitrepos', grid)
        self.AddThreadLocalRootTree(('app', app, 'gitrepos', name))
        return {'NewGitRepo': 'ok'}

    def UpdateGitRepo(self, app, name, doc):
        '''
        Update gitrepo with the data in doc
        '''
        assert app and name and doc
        assert elita.util.type_check.is_string(app)
        assert elita.util.type_check.is_string(name)
        assert elita.util.type_check.is_dictlike(doc)
        assert app in self.root['app']

        self.UpdateObject('gitrepos', {'name': name, 'application': app}, doc)

    def DeleteGitRepo(self, app, name):
        '''
        Delete gitrepo object and root_tree reference
        '''
        assert app and name
        assert elita.util.type_check.is_string(app)
        assert elita.util.type_check.is_string(name)
        assert app in self.root['app']
        assert name in self.root['app'][app]['gitrepos']

        self.mongo_service.rm_roottree(('app', app, 'gitrepos', name))
        self.RmThreadLocalRootTree(('app', app, 'gitrepos', name))
        self.mongo_service.delete('gitrepos', {'name': name, 'application': app})

class DeploymentDataService(GenericChildDataService):
    def NewDeployment(self, app, build_name, environments, groups, servers, gitdeploys, options):
        '''
        Create new deployment object

        @rtype: dict
        '''
        assert app and build_name and options and ((environments and groups) or (servers and gitdeploys))
        assert elita.util.type_check.is_string(app)
        assert elita.util.type_check.is_string(build_name)
        assert elita.util.type_check.is_dictlike(options)
        assert app in self.root['app']
        assert build_name in self.root['app'][app]['builds']
        assert all([not p for p in (environments, groups)]) or all([elita.util.type_check.is_seq(p) for p in (environments, groups)])
        assert all([not p for p in (servers, gitdeploys)]) or all([elita.util.type_check.is_seq(p) for p in (servers, gitdeploys)])
        assert not environments or set(environments).issubset(set(self.deps['ServerDataService'].GetEnvironments()))
        assert not groups or all([g in self.root['app'][app]['groups'] for g in groups])
        assert not servers or all([s in self.root['server'] for s in servers])
        assert not gitdeploys or all([gd in self.root['app'][app]['gitdeploys'] for gd in gitdeploys])


        dpo = Deployment({
            'name': "",
            'application': app,
            'build_name': build_name,
            'environments': environments,
            'groups': groups,
            'servers': servers,
            'gitdeploys': gitdeploys,
            'options': options,
            'status': 'created',
            'job_id': ''
        })

        did = self.mongo_service.create_new('deployments', {}, 'Deployment', dpo.get_doc(), remove_existing=False)
        # we don't know the full deployment 'name' until it's inserted
        name = "{}_{}".format(build_name, str(did))
        self.mongo_service.modify('deployments', {'_id': did}, ('name',), name)
        self.mongo_service.update_roottree(('app', app, 'deployments', name), 'deployments', did)
        self.AddThreadLocalRootTree(('app', app, 'deployments', name))
        return {
            'NewDeployment': {
                'application': app,
                'id': name
            }
        }

    def GetDeployments(self, app, sort=False, with_details=False):
        '''
        Get a list of all deployments for application.

        sort can be either "asc" or "desc" (False indicates no sorting). Sorting always done by created_datetime
        @rtype: list(str)
        '''
        assert app
        assert elita.util.type_check.is_string(app)
        assert app in self.root['app']

        #query from mongo instead of root_tree so we can sort and get datetimes
        deployments = self.mongo_service.get('deployments', {'application': app}, multi=True, empty=True)
        #pymongo does not easily let you sort by generation_time internally, so we have to hack it here
        if sort:
            deployments = sorted(deployments, key=lambda d: d['_id'].generation_time, reverse=(sort == "desc"))
        if with_details:
            return [{"name": doc['name'],
                     "created": doc['_id'].generation_time.isoformat(' '),
                     "status": doc['status']} for doc in deployments]
        else:
            return [doc['name'] for doc in deployments]

    def GetDeployment(self, app, name):
        '''
        Get a specific deployment document
        '''
        assert app and name
        assert app in self.root['app']
        assert name in self.root['app'][app]['deployments']

        doc = self.mongo_service.get('deployments', {'application': app, 'name': name})
        doc['created_datetime'] = doc['_id'].generation_time
        return {k: doc[k] for k in doc if k[0] != '_'}

    def UpdateDeployment(self, app, name, doc):
        '''
        Modify deployment object with the data in doc
        '''
        assert app and name and doc
        assert elita.util.type_check.is_string(app)
        assert elita.util.type_check.is_string(name)
        assert elita.util.type_check.is_dictlike(doc)
        assert app in self.root['app']
        assert name in self.root['app'][app]['deployments']
        
        self.UpdateObject('deployments', {'application': app, 'name': name}, doc)

    def InitializeDeploymentPlan(self, app, name, batches, gitrepos):
        '''
        Create the appropriate structure in the "progress" field of the deployment.

        @type app: str
        @type name: str
        @type batches: list(dict)
        @type gitdeploys: list(str)
        '''
        assert app and name and batches and gitrepos
        assert elita.util.type_check.is_string(app)
        assert elita.util.type_check.is_seq(batches)
        assert all([elita.util.type_check.is_dictlike(b) for b in batches])
        assert elita.util.type_check.is_seq(gitrepos)  # list of all gitdeploys

        for gr in gitrepos:
            doc = {
                'progress': {
                    'phase1': {
                        'gitrepos': {
                            gr: {
                                'progress': 0,
                                'step': 'not started',
                                'changed_files': []
                            }
                        }
                    }
                }
            }
            self.UpdateDeployment(app, name, doc)

        # at a low level, deployment operates in a gitdeploy-centric way
        # but on a human level, a server-centric view is more intuitive, so we generate a list of servers per batch,
        # then the gitdeploys to be deployed on each server:
        # batch 0:
        #   server0:
        #       - gitdeployA
        #           * path: C:\foo\bar
        #           * package: webapplication
        #           * progress: 0%
        #           * state: not started
        #       - gitdeployB
        #           * progress: 10%
        #           * package: webapplication
        #           * progress: 0%
        #           * state: checking out default branch


        doc = {
            'progress': {
                'phase2': {}
            }
        }
        for i, batch in enumerate(batches):
            batch_name = 'batch{}'.format(i)
            doc['progress']['phase2'][batch_name] = {}
            for server in batch['servers']:
                doc['progress']['phase2'][batch_name][server] = {}
                for gitdeploy in batch['gitdeploys']:
                    gddoc = self.deps['GitDataService'].GetGitDeploy(app, gitdeploy)
                    if server in elita.deployment.deploy.determine_deployabe_servers(gddoc['servers'], [server]):
                        doc['progress']['phase2'][batch_name][server][gitdeploy] = {
                            'path': gddoc['location']['path'],
                            'package': gddoc['package'],
                            'progress': 0,
                            'state': 'not started'
                        }

        logging.debug('InitializeDeploymentPlan: doc: {}'.format(doc))
        self.UpdateDeployment(app, name, doc)

    def StartDeployment_Phase(self, app, name, phase):
        '''
        Mark the progress field currently_on as phaseN
        '''
        assert app and name and phase
        assert all([elita.util.type_check.is_string(p) for p in (app, name)])
        assert isinstance(phase, int) and (phase == 1 or phase == 2)
        assert app in self.root['app']
        assert name in self.root['app'][app]['deployments']

        self.UpdateDeployment(app, name, {'progress': {'currently_on': 'phase{}'.format(phase)}})

    def FailDeployment(self, app, name):
        '''
        Mark deployment as failed in event of errors
        '''
        assert app and name
        assert all([elita.util.type_check.is_string(p) for p in (app, name)])
        assert app in self.root['app']
        assert name in self.root['app'][app]['deployments']

        self.UpdateDeployment(app, name, {'progress': {'currently_on': 'failure'}})
        self.UpdateDeployment(app, name, {'status': 'error'})

    def CompleteDeployment(self, app, name):
        '''
        Mark deployment as done
        '''
        assert app and name
        assert all([elita.util.type_check.is_string(p) for p in (app, name)])
        assert app in self.root['app']
        assert name in self.root['app'][app]['deployments']

        self.UpdateDeployment(app, name, {'progress': {'currently_on': 'completed'}})

    def UpdateDeployment_Phase1(self, app, name, gitrepo, progress=None, step=None, changed_files=None):
        '''
        Phase 1 is gitrepo processing: decompressing package to local gitrepo, computing changes, commiting/pushing
        '''
        assert app and name and gitrepo
        assert all([elita.util.type_check.is_string(p) for p in (app, name, gitrepo)])
        assert (isinstance(progress, int) and 0 <= progress <= 100) or not progress
        assert elita.util.type_check.is_string(step) or elita.util.type_check.is_seq(step) or not step
        assert elita.util.type_check.is_optional_seq(changed_files)
        assert app in self.root['app']
        assert name in self.root['app'][app]['deployments']

        progress_dict = dict()
        if progress:
            progress_dict['progress'] = progress
        if step:
            progress_dict['step'] = step
        if changed_files:
            progress_dict['changed_files'] = changed_files

        self.UpdateDeployment(app, name, {'progress': {'phase1': {'gitrepos': {gitrepo: progress_dict}}}})

    def UpdateDeployment_Phase2(self, app, name, gitdeploy, servers, batch, progress=None, state=None):
        '''
        Phase 2 is performing git pulls across all remote servers. Progress is presented per server, but the backend
        deployment is done per gitdeploy (to multiple servers)
        '''
        assert app and name and gitdeploy and servers
        assert all([elita.util.type_check.is_string(p) for p in (app, name, gitdeploy)])
        assert elita.util.type_check.is_seq(servers)
        assert isinstance(batch, int) and batch >= 0
        assert app in self.root['app']
        assert name in self.root['app'][app]['deployments']

        progress_dict = {'batch{}'.format(batch): {}}

        for s in servers:
            progress_dict['batch{}'.format(batch)][s] = dict()
            progress_dict['batch{}'.format(batch)][s][gitdeploy] = dict()
            if progress:
                progress_dict['batch{}'.format(batch)][s][gitdeploy]['progress'] = progress
            if state:
                progress_dict['batch{}'.format(batch)][s][gitdeploy]['state'] = state

        self.UpdateDeployment(app, name, {'progress': {'phase2': progress_dict}})

class KeyDataService(GenericChildDataService):
    def GetKeyPairs(self):
        '''
        Get all extant keypair names
        '''
        return [k for k in self.root['global']['keypairs'] if k[0] != '_']

    def GetKeyPair(self, name):
        '''
        Get keypair doc
        '''
        assert name
        assert elita.util.type_check.is_string(name)
        assert name in self.root['global']['keypairs']

        doc = self.mongo_service.get('keypairs', {'name': name})
        doc['created_datetime'] = doc['_id'].generation_time
        return {k: doc[k] for k in doc if k[0] != '_'} if doc else None

    def NewKeyPair(self, name, attribs, key_type, private_key, public_key):
        '''
        Create new keypair object and root_tree reference
        '''
        assert name and key_type and private_key and public_key
        assert all([elita.util.type_check.is_string(p) for p in (name, key_type, private_key, public_key)])
        assert elita.util.type_check.is_optional_dict(attribs)

        try:
            kp_obj = KeyPair({
                "name": name,
                "attributes": attribs,
                "key_type": key_type,
                "private_key": private_key,
                "public_key": public_key
            })
        except:
            exc_type, exc_obj, tb = sys.exc_info()
            logging.debug("exception: {}, {}".format(exc_type, exc_obj))
            if exc_type == elita_exceptions.InvalidPrivateKey:
                err = "Invalid private key"
            if exc_type == elita_exceptions.InvalidPublicKey:
                err = "Invalid public key"
            if exc_type == elita_exceptions.InvalidKeyPairType:
                err = "Invalid key type"
            else:
                err = "unknown key error"
            return {
                'NewKeyPair': {
                    'status': "error",
                    'message': err
                }
            }
        kpid = self.mongo_service.create_new('keypairs', {'name': name}, 'KeyPair', kp_obj.get_doc())
        self.mongo_service.update_roottree(('global', 'keypairs', name), 'keypairs', kpid)
        self.AddThreadLocalRootTree(('global', 'keypairs', name))
        return {
            'NewKeyPair': {
                'status': 'ok'
            }
        }

    def UpdateKeyPair(self, name, doc):
        '''
        Update key pair object with data in doc
        '''
        assert name and doc
        assert elita.util.type_check.is_string(name)
        assert elita.util.type_check.is_dictlike(doc)
        assert name in self.root['global']['keypairs']

        self.UpdateObject('keypairs', {'name': name}, doc)

    def DeleteKeyPair(self, name):
        '''
        Delete keypair object and root_tree reference
        '''
        assert name
        assert elita.util.type_check.is_string(name)
        assert name in self.root['global']['keypairs']

        self.mongo_service.rm_roottree(('global', 'keypairs', name))
        self.RmThreadLocalRootTree(('global', 'keypairs', name))
        self.mongo_service.delete('keypairs', {'name': name})


class DataService:
    '''
    DataService is an object that holds all the data-layer handling objects. A DataService instance is part of the request
    object and also passed to async jobs, etc. It is the main internal API for data handling.
    '''
    __metaclass__ = elita.util.LoggingMetaClass

    def __init__(self, settings, db, root, job_id=None):
        '''
        @type root: RootTree
        @type db: pymongo.database.Database
        @type settings: pyramid.registry.Registry
        @type job_id: str | None
        '''
        self.settings = settings
        self.db = db
        self.root = root
        self.mongo_service = MongoService(db)

        self.buildsvc = BuildDataService(self.mongo_service, root, settings, job_id=job_id)
        self.usersvc = UserDataService(self.mongo_service, root, settings, job_id=job_id)
        self.appsvc = ApplicationDataService(self.mongo_service, root, settings, job_id=job_id)
        self.jobsvc = JobDataService(self.mongo_service, root, settings, job_id=job_id)
        self.serversvc = ServerDataService(self.mongo_service, root, settings, job_id=job_id)
        self.gitsvc = GitDataService(self.mongo_service, root, settings, job_id=job_id)
        self.keysvc = KeyDataService(self.mongo_service, root, settings, job_id=job_id)
        self.deploysvc = DeploymentDataService(self.mongo_service, root, settings, job_id=job_id)
        self.actionsvc = ActionService(self)
        self.groupsvc = GroupDataService(self.mongo_service, root, settings, job_id=job_id)
        self.pmsvc = PackageMapDataService(self.mongo_service, root, settings, job_id=job_id)

        #cross-dependencies between child dataservice objects above
        self.appsvc.populate_dependencies({
            'ServerDataService': self.serversvc,
            'GroupDataService': self.groupsvc,
            'GitDataService': self.gitsvc
        })
        self.groupsvc.populate_dependencies({
            'ServerDataService': self.serversvc,
            'GitDataService': self.gitsvc
        })
        self.jobsvc.populate_dependencies({
            'ActionService': self.actionsvc,
            'ApplicationDataService': self.appsvc
        })
        self.deploysvc.populate_dependencies({
            'ServerDataService': self.serversvc,
            'GroupDataService': self.groupsvc,
            'GitDataService': self.gitsvc
        })

        #load all plugins and register actions/hooks
        self.actionsvc.register()

        #passed in if this is part of an async job
        self.job_id = job_id
        #super ugly below - only exists for plugin access
        if job_id is not None:
            self.salt_controller = salt_control.SaltController(self)
            self.remote_controller = salt_control.RemoteCommands(self.salt_controller)

    def GetAppKeys(self, app):
        return [k for k in self.root['app'][app] if k[0] != '_']

    def GetGlobalKeys(self):
        return [k for k in self.root['global'] if k[0] != '_']

class SupportedFileType:
    TarGz = 'tar.gz'
    TarBz2 = 'tar.bz2'
    Zip = 'zip'
    types = [TarBz2, TarGz, Zip]


class GenericDataModel:
    __metaclass__ = elita.util.LoggingMetaClass

    default_values = {}

    def __init__(self, doc):
        self.set_defaults()
        if doc is not None:
            for k in doc:
                if k == '_id':
                    self.created_datetime = doc['_id'].generation_time
                elif k[0] != '_':
                    self.set_data(k, doc[k])
        self.init_hook()

    def init_hook(self):
        pass

    def set_defaults(self):
        for k in self.default_values:
            if not hasattr(self, k):
                setattr(self, k, self.default_values[k])

    def set_data(self, key, value):
        self._set_data(key, value)

    def _set_data(self, key, value):
        setattr(self, key, value)

    def update_values(self, doc):
        for k in doc:
            if hasattr(self, k):
                setattr(self, k, doc[k])

    def get_doc(self):
        return {k: getattr(self, k) for k in self.default_values}


class KeyPair(GenericDataModel):
    default_values = {
        "name": None,
        "attributes": dict(),
        "key_type": None,  # git | salt
        "private_key": None,
        "public_key": None
    }

    def init_hook(self):
        if self.key_type not in ("git", "salt"):
            raise elita_exceptions.InvalidKeyPairType
        kp = keypair.KeyPair(self.private_key, self.public_key)
        try:
            kp.verify_public()
        except:
            raise elita_exceptions.InvalidPublicKey
        try:
            kp.verify_private()
        except:
            raise elita_exceptions.InvalidPrivateKey

class Server(GenericDataModel):
    default_values = {
        "name": None,
        "server_type": None,
        "status": None,
        "environment": None,
        "attributes": dict(),
        "gitdeploys": [bson.DBRef("", None)],
        "salt_key": bson.DBRef("", None)
    }

class Environment(GenericDataModel):
    default_values = {
        "environments": None
    }

class GitProvider(GenericDataModel):
    default_values = {
        'name': None,
        'type': None,
        'auth': {
            'username': None,
            'password': None
        }
    }

class GitRepo(GenericDataModel):
    default_values = {
        'name': None,
        'application': None,
        'last_build': None,
        'uri': None,
        'keypair': bson.DBRef("", None),
        'gitprovider': bson.DBRef("", None)
    }

class GitDeploy(GenericDataModel):
    default_values = {
        'name': None,
        'application': None,
        'package': 'master',
        'servers': list(),
        'attributes': dict(),
        'deployed_build': None,
        'options': {
            'favor': 'ours',
            'ignore-whitespace': 'true',
            'gitignore': list()
        },
        'actions': {
            'prepull': [],
            'postpull': []
        },
        'location': {
            'path': None,
            'git_repo': bson.DBRef("", None),
            'default_branch': None,
            'current_branch': None,
            'current_rev': None
        }
    }

class Deployment(GenericDataModel):
    '''
    There are two "phases" to a deployment:
        - Phase 1: processing of gitdeploys (applying packages, adding/committing/pushing to gitprovider, determining
            changes)
        - Phase 2: Performing salt states/git pull on target machines. Can be broken up into an arbitrary number of
            batches
    '''
    default_values = {
        'name': None,  # "name" for consistency w/ other models, even though it's really id
        'application': None,
        'build_name': None,
        'environments': None,
        'groups': None,
        'servers': None,
        'gitdeploys': None,
        'options': dict(),  # pauses, divisor
        'status': None,
        'commits': dict(),    # { gitrepo_name: commit_hash }
        'progress': {
            'currently_on': None,
            'phase1': {
                'gitrepos': dict()
            },
            'phase2': dict()
        },
        'job_id': None
    }

class Build(GenericDataModel):
    default_values = {
        'build_name': None,
        'app_name': None,
        'files': list(),
        'stored': False,
        'master_file': None,
        'packages': dict(),
        'attributes': dict()
    }


class Action:
    def __init__(self, app_name, action_name, params, job_datasvc):
        self.app_name = app_name
        self.action_name = action_name
        self.params = params
        self.job_datasvc = job_datasvc

    def execute(self, params):
        return self.job_datasvc.ExecuteAction(self.app_name, self.action_name, params)

    def details(self):
        return self.job_datasvc.GetAction(self.app_name, self.action_name)

class Application(GenericDataModel):
    default_values = {
        'app_name': None
    }

DEFAULT_ADMIN_USERNAME = 'admin'
DEFAULT_ADMIN_PASSWORD = 'elita'
DEFAULT_ADMIN_PERMISSIONS = {
    'apps': {
        '*': 'read/write',
        '_global': 'read/write'
    },
    'actions': {
        '*': {
            '*': 'execute'
        }
    },
    'servers': ['*']
}

class User(GenericDataModel):
    default_values = {
        'username': None,
        'password': None,
        'hashed_pw': None,
        'salt': None,
        'attributes': dict(),
        'permissions': {
            'apps': {},
            'actions': {},
            'servers': []
        }
    }

    def init_hook(self):
        assert self.hashed_pw or self.password
        if self.password:   # new or changed password
            self.salt = base64.urlsafe_b64encode(uuid.uuid4().bytes)
            self.hashed_pw = self.hash_pw(self.password)
            self.password = None

    def hash_pw(self, pw):
        assert pw is not None
        return base64.urlsafe_b64encode(hashlib.sha512(pw + self.salt).hexdigest())

    def validate_password(self, pw):
        return self.hashed_pw == self.hash_pw(pw)


#dummy model. Values are computed for each request
class UserPermissions(GenericDataModel):
    default_values = {
        'username': None,
        'applications': None,
        'actions': None,
        'servers': None
    }

class Token(GenericDataModel):
    default_values = {
        'username': None,
        'token': None
    }

    def init_hook(self):
        self.token = base64.urlsafe_b64encode(hashlib.sha256(uuid.uuid4().bytes).hexdigest())[:-2] \
            if self.token is None else self.token

class Job(GenericDataModel):
    default_values = {
        'status': 'running',
        'name': None,
        'job_type': None,
        'data': None,
        'completed_datetime': None,
        'duration_in_seconds': None,
        'attributes': None,
        'job_id': None
    }

    def init_hook(self):
        self.job_id = uuid.uuid4() if self.job_id is None else self.job_id

class Group(GenericDataModel):
    default_values = {
        'application': None,
        'name': None,
        'description': None,
        'gitdeploys': list(),
        'attributes': dict(),
        'rolling_deploy': False
    }

class PackageMap(GenericDataModel):
    default_values = {
        'application': None,
        'name': None,
        'attributes': dict(),
        'packages': dict()
    }
    # Example packages field:
    # packages = {
    #     'package_name': [{
    #             'patterns': ["foo/bar/*"],
    #             'prefix': "foobar/",
    #             'remove_prefix': "bar/"
    #     }]
    # }

class GenericContainer:
    __metaclass__ = elita.util.LoggingMetaClass

    def __init__(self, doc):
        self.name = doc['name']
        self.parent = doc['parent']
        self.created_datetime = doc['_id'].generation_time if '_id' in doc else None


class BuildContainer(GenericContainer):
    pass

class ActionContainer(GenericContainer):
    pass

class UserContainer(GenericContainer):
    pass

class TokenContainer(GenericContainer):
    pass

class ApplicationContainer(GenericContainer):
    pass

class GlobalContainer(GenericContainer):
    pass

class JobContainer(GenericContainer):
    pass

class ServerContainer(GenericContainer):
    pass

class GitProviderContainer(GenericContainer):
    pass

class GitRepoContainer(GenericContainer):
    pass

class GitDeployContainer(GenericContainer):
    pass

class DeploymentContainer(GenericContainer):
    pass

class KeyPairContainer(GenericContainer):
    pass

class GroupContainer(GenericContainer):
    pass

class PackageMapContainer(GenericContainer):
    pass

class Root(GenericContainer):
    pass

class RootTree(collections.MutableMapping):

    def __init__(self, db, updater, tree, doc, *args, **kwargs):
        self.pp = pprint.PrettyPrinter(indent=4)
        self.db = db
        self.tree = tree
        self.doc = doc
        self.updater = updater

    def is_action(self):
        return self.doc and self.doc['_class'] == 'ActionContainer'

    def __getitem__(self, key):
        key = self.__keytransform__(key)
        if self.is_action():
            return self.tree[key]
        if key in self.tree:
            if key == '_doc':
                return self.tree[key]
            doc = self.db.dereference(self.tree[key]['_doc'])
            if doc is None:
                logging.debug("RootTree: __getitem__: {}: doc is None: KeyError".format(key))
                raise KeyError
            return RootTree(self.db, self.updater, self.tree[key], doc)
        else:
            logging.debug("RootTree: __getitem__: {}: key not in self.tree: KeyError".format(key))
            raise KeyError

    def __setitem__(self, key, value):
        self.tree[key] = value
        if not self.is_action():     # dynamically populated each request
            pass
            #self.updater.update()

    def __delitem__(self, key):
        del self.tree[key]
        #self.updater.update()
        return

    def __iter__(self):
        return iter(self.tree)

    def __len__(self):
        return len(self.tree)

    def __keytransform__(self, key):
        return key

class RootTreeUpdater:

    def __init__(self, tree, db):
        self.tree = tree
        self.db = db

    def clean_actions(self):
        #actions can't be serialized into mongo
        for a in self.tree['app']:
            actions = list()
            if a[0] != '_':
                if "actions" in self.tree['app'][a]:
                    for ac in self.tree['app'][a]['actions']:
                        if ac[0] != '_':
                            actions.append(ac)
                    for action in actions:
                        del self.tree['app'][a]['actions'][action]

    def update(self):
        self.clean_actions()
        root_tree = self.db['root_tree'].find_one()
        self.db['root_tree'].update({"_id": root_tree['_id']}, self.tree)

class DataValidator:
    '''
    Responsible for:
        - Creating proper data model/root_tree on first run
        - Validating that data isn't broken/inconsistent
        - Running migrations between versions
    '''
    __metaclass__ = util.LoggingMetaClass

    def __init__(self, settings, root, db):
        '''
        @type settings: pyramid.registry.Registry
        @type root: dict
        @type db: pymongo.database.Database
        '''
        # root is optional if this is the first time elita runs
        assert settings and db
        self.settings = settings
        self.root = root
        self.db = db

    def run(self):
        # order is very significant
        logging.debug("running")
        self.check_root()
        self.check_toplevel()
        self.check_global()
        self.check_root_references()
        self.check_users()
        self.check_user_permissions()
        self.check_doc_consistency()
        self.check_containers()
        self.check_apps()
        self.check_jobs()
        self.check_deployments()
        self.check_servers()
        self.check_gitdeploys()
        self.check_gitrepos()
        self.check_groups()
        self.SaveRoot()

    def SaveRoot(self):
        logging.debug('saving root_tree')
        self.db['root_tree'].save(self.root)

    def NewContainer(self, class_name, name, parent):
        cdoc = self.db['containers'].insert({'_class': class_name,
                                         'name': name,
                                         'parent': parent})
        return bson.DBRef('containers', cdoc)

    def check_doc_consistency(self):
        #enforce collection rules and verify connectivity b/w objects and root tree
        collects = [
            'applications',
            'builds',
            'gitproviders',
            'gitrepos',
            'gitdeploys',
            'jobs',
            'tokens',
            'users',
            'servers',
            'groups',
            'packagemaps'
        ]
        for c in collects:
            rm_docs = list()
            for d in self.db[c].find():
                if '_class' not in d:
                    logging.warning("class specification not found in object {} from collection {}"
                    .format(d['_id'], c))
                    logging.debug("...deleting malformed object")
                    self.db[c].remove({'_id': d['_id']})
                if c == 'applications':
                    if d['app_name'] not in self.root['app']:
                        logging.warning("orphan application object: '{}', adding to root tree".format(d['app_name']))
                        self.root['app'][d['app_name']] = {'_doc': bson.DBRef('applications', d['_id'])}
                    if self.root['app'][d['app_name']]['_doc'].id != d['_id']:
                        logging.warning("unlinked application document found, deleting")
                        rm_docs.append(d['_id'])
                if c == 'builds':
                    if d['build_name'] not in self.root['app'][d['app_name']]['builds']:
                        logging.warning("orphan build object: '{}/{}', adding to root tree".format(d['app_name'], d['build_name']))
                        self.root['app'][d['app_name']]['builds'][d['build_name']] = {'_doc': bson.DBRef('builds', d['_id'])}
                    if self.root['app'][d['app_name']]['builds'][d['build_name']]['_doc'].id != d['_id']:
                        logging.warning("unlinked build document found, deleting")
                        rm_docs.append(d['_id'])
                if c == 'gitproviders':
                    if d['name'] not in self.root['global']['gitproviders']:
                        logging.warning("orphan gitprovider object: '{}', adding to root tree".format(d['name']))
                        self.root['global']['gitproviders'][d['name']] = {'_doc': bson.DBRef('gitproviders', d['_id'])}
                    if self.root['global']['gitproviders'][d['name']]['_doc'].id != d['_id']:
                        logging.warning("unlinked gitprovider document found, deleting")
                        rm_docs.append(d['_id'])
                if c == 'gitrepos':
                    if d['name'] not in self.root['app'][d['application']]['gitrepos']:
                        logging.warning("orphan gitrepo object: '{}/{}', adding to root tree".format(d['application'], d['name']))
                        self.root['app'][d['application']]['gitrepos'][d['name']] = {'_doc': bson.DBRef('gitrepos', d['_id'])}
                    if self.root['app'][d['application']]['gitrepos'][d['name']]['_doc'].id != d['_id']:
                        logging.warning("unlinked gitrepo document found, deleting")
                        rm_docs.append(d['_id'])
                if c == 'gitdeploys':
                    if d['name'] not in self.root['app'][d['application']]['gitdeploys']:
                        logging.warning("orphan gitdeploy object: '{}/{}', adding to root tree".format(d['application'], d['name']))
                        self.root['app'][d['application']]['gitdeploys'][d['name']] = {'_doc': bson.DBRef('gitdeploys', d['_id'])}
                    if self.root['app'][d['application']]['gitdeploys'][d['name']]['_doc'].id != d['_id']:
                        logging.warning("unlinked gitdeploy document found, deleting")
                        rm_docs.append(d['_id'])
                if c == 'jobs':
                    if d['job_id'] not in self.root['job']:
                        logging.warning("orphan job object: '{}', adding to root tree".format(d['job_id']))
                        self.root['job'][d['job_id']] = {'_doc': bson.DBRef('jobs', d['_id'])}
                    #we don't really care about unlinked jobs
                if c == 'tokens':
                    if d['token'] not in self.root['global']['tokens']:
                        logging.warning("orphan token object: '{}', adding to root tree".format(d['token']))
                        self.root['global']['tokens'][d['token']] = {'_doc': bson.DBRef('tokens', d['_id'])}
                    if self.root['global']['tokens'][d['token']]['_doc'].id != d['_id']:
                        logging.warning("unlinked token document found, deleting")
                        rm_docs.append(d['_id'])
                if c == 'users':
                    if d['username'] not in self.root['global']['users']:
                        logging.warning("orphan user object: '{}', adding to root tree".format(d['username']))
                        self.root['global']['users'][d['username']] = {'_doc': bson.DBRef('users', d['_id'])}
                    if self.root['global']['users'][d['username']]['_doc'].id != d['_id']:
                        logging.warning("unlinked user document found ({}), deleting".format(d['username']))
                        rm_docs.append(d['_id'])
                if c == 'servers':
                    if d['name'] not in self.root['server']:
                        logging.warning("orphan server object: '{}', adding to root tree".format(d['name']))
                        self.root['server'][d['name']] = {'_doc': bson.DBRef('servers', d['_id'])}
                    if self.root['server'][d['name']]['_doc'].id != d['_id']:
                        logging.warning("unlinked server document found, deleting")
                        rm_docs.append(d['_id'])
                if c == 'groups':
                    if d['name'] not in self.root['app'][d['application']]['groups']:
                        logging.warning("orphan group object: '{}', adding to root tree".format(d['name']))
                        self.root['app'][d['application']]['groups'][d['name']] = {'_doc': bson.DBRef('groups', d['_id'])}
                    if self.root['app'][d['application']]['groups'][d['name']]['_doc'].id != d['_id']:
                        logging.warning("unlinked group document found ({}), deleting".format(d['name']))
                        rm_docs.append(d['_id'])
                if c == 'packagemaps':
                    if d['name'] not in self.root['app'][d['application']]['packagemaps']:
                        logging.warning("orphan packagemap object: '{}', adding to root tree".format(d['name']))
                        self.root['app'][d['application']]['packagemaps'][d['name']] = {'_doc': bson.DBRef('packagemaps', d['_id'])}
                    if self.root['app'][d['application']]['packagemaps'][d['name']]['_doc'].id != d['_id']:
                        logging.warning("unlinked packagemap document found ({}), deleting".format(d['name']))
                        rm_docs.append(d['_id'])
            for id in rm_docs:
                self.db[c].remove({'_id': id})

    @staticmethod
    def check_root_refs(db, obj):
        '''
        Recurse into obj, find DBRefs, check if they are valid
        '''
        for k in obj:
            if k == '_doc':
                if not db.dereference(obj[k]):
                    logging.warning("Invalid dbref found! {}".format(obj[k].collection))
            elif k[0] != '_' and not isinstance(obj[k], bson.ObjectId):
                DataValidator.check_root_refs(db, obj[k])

    def check_root_references(self):
        '''
        Check that every DBRef in the root tree points to a valid document
        '''
        DataValidator.check_root_refs(self.db, self.root)

    def check_global(self):
        global_levels = {
            'users': {
                'class': "UserContainer"
            },
            'tokens': {
                'class': "TokenContainer"
            },
            'gitproviders': {
                'class': "GitProviderContainer"
            },
            'keypairs': {
                'class': "KeyPairContainer"
            }
        }
        for l in global_levels:
            if l not in self.root['global']:
                logging.warning("'{}' not found under global".format(l))
                self.root['global'][l] = dict()
                self.root['global'][l]['_doc'] = self.NewContainer(global_levels[l]['class'], l, "")
        if DEFAULT_ADMIN_USERNAME not in self.root['global']['users']:
            logging.warning("admin user not found")
            users = [u for u in self.db['users'].find()]
            if DEFAULT_ADMIN_USERNAME in [u['username'] for u in users]:
                admin = self.db['users'].find_one({'username': DEFAULT_ADMIN_USERNAME})
                assert admin
                self.root['global']['users'][DEFAULT_ADMIN_USERNAME] = {
                    '_doc': bson.DBRef('users', admin['_id'])
                }

    def check_users(self):
        for u in self.root['global']['users']:
            if u != '_doc':
                if "permissions" not in self.root['global']['users'][u]:
                    logging.warning("permissions container object not found under user {} in root tree; "
                                        "fixing".format(u))
                    pid = self.db['userpermissions'].insert({
                            "_class": "UserPermissions",
                            "username": u,
                            "applications": list(),
                            "actions": dict(),
                            "servers": list()
                    })
                    self.root['global']['users'][u]['permissions'] = {
                        '_doc': bson.DBRef('userpermissions', pid)
                    }
        users = [u for u in self.db['users'].find()]
        for u in users:
            if 'username' not in u:
                logging.warning("user found without username property; fixing {}".format(u['_id']))
                if 'name' in u:
                    u['username'] = u['name']
                    self.db['users'].save(u)
                else:
                    logging.warning("...couldn't fix user because name field not found!")
        if DEFAULT_ADMIN_USERNAME not in [u['username'] for u in users]:
            logging.warning("admin user document not found; creating")
            userobj = User({
                'username': DEFAULT_ADMIN_USERNAME,
                'permissions': DEFAULT_ADMIN_PERMISSIONS,
                'password': DEFAULT_ADMIN_PASSWORD
            })
            doc = userobj.get_doc()
            doc['_class'] = 'User'
            uid = self.db['users'].insert(doc)
            pid = self.db['userpermissions'].insert({
                "_class": "UserPermissions",
                "username": DEFAULT_ADMIN_USERNAME,
                "applications": list(),
                "actions": dict(),
                "servers": list()
            })
            self.root['global']['users'][DEFAULT_ADMIN_USERNAME] = {
                '_doc': bson.DBRef('users', uid),
                'permissions': {
                    '_doc': bson.DBRef('userpermissions', pid)
                }
            }

    def check_user_permissions(self):
        for u in self.db['users'].find():
            if 'permissions' not in u:
                logging.warning("permissions object not found for user {}; fixing with empty obj"
                              .format(u['username']))
                u['permissions'] = {
                    'apps': {},
                    'actions': {},
                    'servers': []
                }
            for o in ('apps', 'actions', 'servers'):
                if o not in u['permissions']:
                    logging.warning("{} not found in permissions object for user {}; fixing with empty "
                                        "obj".format(o, u['username']))
                    u['permissions'][o] = list() if o == 'servers' else dict()
            if not isinstance(u['permissions']['servers'], list):
                logging.warning("servers key under permissions object for user {} is not a list; fixing"
                              .format(u['username']))
                u['permissions']['servers'] = list(u['permissions']['servers'])
            for o in ('apps', 'actions'):
                if not isinstance(u['permissions'][o], dict):
                    logging.warning("{} key under permissions object for user {} is not a dict; invalid "
                                        "so replacing with empty dict".format(o, u['username']))
                    u['permissions'][o] = dict()
            self.db['users'].save(u)

    def check_jobs(self):
        djs = list()
        for j in self.root['job']:
            if j[0] != '_':
                if self.db.dereference(self.root['job'][j]['_doc']) is None:
                    logging.warning("found dangling job ref in root tree: {}; deleting".format(j))
                    djs.append(j)
        for j in djs:
            del self.root['job'][j]
        job_fixlist = list()
        for doc in self.db['jobs'].find():
            if doc['job_id'] not in self.root['job']:
                job_id = doc['job_id']
                logging.warning("found orphan job: {}; adding to root tree".format(doc['job_id']))
                self.root['job'][job_id] = {'_doc': bson.DBRef('jobs', doc['_id'])}
            for k in ('job_type', 'data', 'name'):
                if k not in doc:
                    logging.warning("found job without {}: {}; adding blank".format(k, doc['job_id']))
                    doc[k] = ""
                    job_fixlist.append(doc)
        for d in job_fixlist:
            self.db['jobs'].save(d)

    def check_apps(self):
        app_sublevels = {
            'builds': {
                'class': "BuildContainer"
            },
            'actions': {
                'class': "ActionContainer"
            },
            'gitrepos': {
                'class': "GitRepoContainer"
            },
            'gitdeploys': {
                'class': "GitDeployContainer"
            },
            'deployments': {
                'class': "DeploymentContainer"
            },
            'groups': {
                'class': "GroupContainer"
            },
            'packagemaps': {
                'class': "PackageMapContainer"
            }
        }
        for a in self.root['app']:
            if a[0] != '_':
                if 'action' in self.root['app'][a]:
                    logging.warning("found old 'action' container under app {}; deleting".format(a))
                    del self.root['app'][a]['action']
                for sl in app_sublevels:
                    if sl not in self.root['app'][a]:
                        logging.warning("'{}' not found under {}".format(sl, a))
                        self.root['app'][a][sl] = dict()
                        self.root['app'][a][sl]['_doc'] = self.NewContainer(app_sublevels[sl]['class'], sl, a)
                    d_o = list()
                    for o in self.root['app'][a][sl]:
                        if o[0] != '_' and '_doc' in self.root['app'][a][sl][o]:
                            if self.db.dereference(self.root['app'][a][sl][o]['_doc']) is None:
                                logging.warning("dangling {} reference '{}' in app {}; deleting".format(sl, o, a))
                                d_o.append(o)
                    for d in d_o:
                        del self.root['app'][a][sl][d]

    def check_servers(self):
        ds = list()
        for s in self.root['server']:
            if s[0] != '_':
                if self.db.dereference(self.root['server'][s]['_doc']) is None:
                    logging.warning("found dangling server ref in root tree: {}; deleting".format(s))
                    ds.append(s)
        for s in ds:
            del self.root['server'][s]
        update_list = list()
        for d in self.db['servers'].find():
            update = False
            if 'environment' not in d:
                logging.warning("environment field not found for server {}; fixing".format(d['_id']))
                d['environment'] = ""
                update = True
            if 'status' not in d:
                logging.warning("status field not found for server {}; fixing".format(d['_id']))
                d['status'] = "ok"
                update = True
            if update:
                update_list.append(d)
        for d in update_list:
            self.db['servers'].save(d)

    def check_deployments(self):
        update_list = list()
        for d in self.db['deployments'].find():
            if 'server_specs' in d:
                logging.warning("found deployment with old-style server_specs object: {}; fixing".format(
                    d['_id']))
                d['deployment'] = {
                    'servers': d['server_specs']['spec'],
                    'gitdeploys': d['server_specs']['gitdeploys']
                }
                update_list.append(d)
            if 'job_id' not in d:
                logging.warning("found deployment without job_id: {}; fixing with blank job".format(d[
                    '_id']))
                d['job_id'] = ""
                update_list.append(d)
            if 'results' in d:
                logging.warning("found deployment with old-style results field: {}; removing"
                              .format(d['_id']))
                update_list.append(d)
            if 'status' not in d:
                logging.warning("found deployment without status: {}; fixing with blank".format(d['_id']))
                d['status'] = ""
                update_list.append(d)
            if 'progress' not in d:
                logging.warning("found deployment without progress: {}; fixing with empty dict".format(d['_id']))
                d['progress'] = dict()
                update_list.append(d)
        for d in update_list:
            if 'server_specs' in d:
                del d['server_specs']
            if 'results' in d:
                del d['results']
            self.db['deployments'].save(d)

        update_list = list()
        for a in self.root['app']:
            if a[0] != '_':
                for d in self.root['app'][a]['deployments']:
                    if d[0] != '_':
                        doc = self.db.dereference(self.root['app'][a]['deployments'][d]['_doc'])
                        assert doc is not None
                        if 'name' not in doc:
                            logging.warning("name not found under deployment {}; fixing".format(d))
                            doc['name'] = d
                            update_list.append(doc)
        if len(update_list) > 0:
            for d in update_list:
                self.db['deployments'].save(d)

    def check_gitdeploys(self):
        dlist = list()
        fixlist = list()
        for d in self.db['gitdeploys'].find():
            delete = False
            for k in ('application', 'name', 'package', 'location'):
                if k not in d:
                    logging.warning("mandatory key '{}' not found under gitdeploy {}; removing".
                                  format(k, d['_id']))
                    dlist.append(d['_id'])
                    delete = True
            fix = False
            if 'attributes' not in d and not delete:
                logging.warning("attributes not found under gitdeploy {}; fixing".format(d['_id']))
                d['attributes'] = dict()
                fix = True
            if 'actions' not in d and not delete:
                logging.warning("actions not found under gitdeploy {}; fixing".format(d['_id']))
                d['actions'] = {
                    'prepull': dict(),
                    'postpull': dict()
                }
                fix = True
            if 'servers' not in d and not delete:
                logging.warning("servers not found under gitdeploy {}; fixing".format(d['_id']))
                d['servers'] = list()
                fix = True
            if 'server' in d:
                logging.warning("gitdeploy found with obsolete server field {}; removing".format(d['_id']))
                del d['server']
                fix = True
            if len([x for x, y in collections.Counter(d['servers']).items() if y > 1]) > 0 and not delete:
                logging.warning("duplicate server entries found in gitdeploy {}; fixing".format(d['_id']))
                d['servers'] = list(set(tuple(d['servers'])))
                fix = True
            if 'deployed_build' not in d:
                logging.warning("deployed_build not found in gitdeploy {}".format(d['_id']))
                gr_doc = self.db.dereference(d['location']['gitrepo'])
                if not gr_doc or 'last_build' not in gr_doc:
                    logging.debug("...ERROR: referenced gitrepo not found! Aborting fix...")
                else:
                    d['deployed_build'] = gr_doc['last_build']
                    fix = True
            if fix:
                fixlist.append(d)
        for f in fixlist:
            self.db['gitdeploys'].find_and_modify(query={'_id': f['_id']}, update=f, upsert=True,
                                                  new=True)
        for dl in dlist:
            self.db['gitdeploys'].remove({'_id': dl})

    def check_gitrepos(self):
        fixlist = list()
        for d in self.db['gitrepos'].find():
            if 'uri' in d:
                if d['uri'] is not None and ':' in d['uri']:
                    logging.warning("found gitrepo URI with ':'; replacing with '/' ({})".format(d['name']))
                    d['uri'] = d['uri'].replace(':', '/')
                    fixlist.append(d)
            else:
                logging.warning("found gitrepo without URI ({}); adding empty field".format(d['name']))
                d['uri'] = ""
                fixlist.append(d)
            if 'last_build' not in d:
                logging.warning("found gitrepo without last_build: {}; adding empty field".format(d['name']))
                d['last_build'] = None
                fixlist.append(d)
        for d in fixlist:
            self.db['gitrepos'].save(d)

    def check_groups(self):
        fixlist = list()
        for d in self.db['groups'].find():
            fix = False
            if 'description' not in d:
                logging.warning("found group without description: {}; fixing".format(d['name']))
                d['description'] = None
                fix = True
            if 'servers' in d:
                logging.warning("found group with explicit server list: {}; removing".format(d['name']))
                del d['servers']
                fix = True
            if 'rolling_deploy' not in d:
                logging.warning("found group without rolling_deploy flag: {}; fixing".format(d['name']))
                d['rolling_deploy'] = False
                fix = True
            if fix:
                fixlist.append(d)
        for d in fixlist:
            self.db['groups'].save(d)

    def check_toplevel(self):
        top_levels = {
            'app': {
                'class': "ApplicationContainer"
            },
            'global': {
                'class': "GlobalContainer"
            },
            'job': {
                'class': 'JobContainer'
            },
            'server': {
                'class': 'ServerContainer'
            }
        }
        for tl in top_levels:
            if tl not in self.root:
                logging.warning("'{}' not found under root".format(tl))
                self.root[tl] = dict()
                self.root[tl]['_doc'] = self.NewContainer(top_levels[tl]['class'], tl, "")
            if tl == 'server':
                if "environments" not in self.root[tl] or isinstance(self.root[tl]['environments'], bson.DBRef):
                    logging.warning("environments endpoint not found under servers container; fixing")
                    eid = self.db['environments'].insert({
                        '_class': "Environment",
                        'environments': ""
                    })
                    self.root[tl]['environments'] = {
                        "_doc": bson.DBRef("environments", eid)
                    }

    def check_containers(self):
        update_list = list()
        for c in self.db['containers'].find():
            if c['_class'] == 'AppContainer':
                logging.warning("found Application container with incorrect 'AppContainer' class; fixing")
                c['_class'] = 'ApplicationContainer'
                update_list.append(c)
        for c in update_list:
            self.db['containers'].save(c)

    def check_root(self):
        if not self.root:
            logging.warning("no root_tree found!")
        self.root = dict() if not self.root else self.root
        if '_class' in self.root:
            logging.warning("'_class' found in base of root; deleting")
            del self.root['_class']
        if '_doc' not in self.root:
            logging.warning("'_doc' not found in base of root")
            self.root['_doc'] = self.NewContainer('Root', 'Root', '')
        rt_list = [d for d in self.db['root_tree'].find()]
        if len(rt_list) > 1:
            logging.warning("duplicate root_tree docs found! Removing all but the first")
            for d in rt_list[1:]:
                self.db['root_tree'].remove({'_id': d['_id']})





#requires package_map_data/build-123.*

import os, zipfile, tarfile, os
from elita.builds import PackageMapper, SupportedFileType

BUILD_BASENAME = "package_map_data/build-123"
BUILD_ZIP = "{}.zip".format(BUILD_BASENAME)
BUILD_TGZ = "{}.tar.gz".format(BUILD_BASENAME)
BUILD_TBZ2 = "{}.tar.bz2".format(BUILD_BASENAME)

package_map = {
    "configuration": {
        "patterns": ["build-123/config/**/*"],
        "remove_prefix": "build-123/"
    },
    "binaries": {
        "patterns": ["build-123/bin/**/*"]
    },
    "libraries": {
        "patterns": ["build-123/bin/lib/**/*"]
    },
    "assets": {
        "patterns": ["build-123/assets/**/*"],
        "prefix": "web/static/"
    }
}

def test_packagemap_creates_expected_packages_zip():
    '''
    Test that PackageMapper creates expected packages from master zip
    '''

    cwd = os.getcwd()
    output_dir = "{}/package_map_data/output".format(cwd)
    pm = PackageMapper(BUILD_ZIP, SupportedFileType.Zip, SupportedFileType.Zip, output_dir, package_map)
    pm.apply()
    pm.cleanup()

    expected_output = ['assets.zip', 'configuration.zip', 'binaries.zip', 'libraries.zip']
    files = os.listdir(output_dir)
    assert set(expected_output) == set([f for f in files if f[0] != '.'])   # filter out .keep

    packages = ['configuration.zip', 'binaries.zip', 'libraries.zip', 'assets.zip']
    expected_members = {
        'configuration.zip': ['config/file.yaml', 'config/file.xml', 'config/file.txt'],
        'binaries.zip': ['build-123/bin/lib/', 'build-123/bin/widgets', 'build-123/bin/lib/libfoo.so'],
        'libraries.zip': ['build-123/bin/lib/libfoo.so'],
        'assets.zip': ['web/static/build-123/assets/css/', 'web/static/build-123/assets/css/style.css',
                       'web/static/build-123/assets/main.html', 'web/static/build-123/assets/img/',
                       'web/static/build-123/assets/img/banner.png']
    }

    for p in packages:
        pname = "{}/{}".format(output_dir, p)
        with zipfile.ZipFile(pname, 'r') as zf:
            assert set(zf.namelist()) == set(expected_members[p])
        os.unlink(pname)

def test_packagemap_creates_expected_packages_tgz():
    '''
    Test that PackageMapper creates expected packages from master .tar.gz
    '''

    cwd = os.getcwd()
    output_dir = "{}/package_map_data/output".format(cwd)
    pm = PackageMapper(BUILD_TGZ, SupportedFileType.TarGz, SupportedFileType.TarGz, output_dir, package_map)
    pm.apply()
    pm.cleanup()

    expected_output = ['assets.tar.gz', 'configuration.tar.gz', 'binaries.tar.gz', 'libraries.tar.gz']
    files = os.listdir(output_dir)
    assert set(expected_output) == set([f for f in files if f[0] != '.'])   # filter out .keep

    packages = ['configuration.tar.gz', 'binaries.tar.gz', 'libraries.tar.gz', 'assets.tar.gz']
    # note trailing / removed from empty directories (per tarfile output)
    expected_members = {
        'configuration.tar.gz': ['config/file.yaml', 'config/file.xml', 'config/file.txt'],
        'binaries.tar.gz': ['build-123/bin/lib', 'build-123/bin/widgets', 'build-123/bin/lib/libfoo.so'],
        'libraries.tar.gz': ['build-123/bin/lib/libfoo.so'],
        'assets.tar.gz': ['web/static/build-123/assets/css', 'web/static/build-123/assets/css/style.css',
                       'web/static/build-123/assets/main.html', 'web/static/build-123/assets/img',
                       'web/static/build-123/assets/img/banner.png']
    }

    for p in packages:
        pname = "{}/{}".format(output_dir, p)
        with tarfile.open(pname, 'r:gz') as tf:
            assert set(tf.getnames()) == set(expected_members[p])
        os.unlink(pname)


def test_packagemap_creates_expected_packages_tbz2():
    '''
    Test that PackageMapper creates expected packages from master .tar.bz2
    '''

    cwd = os.getcwd()
    output_dir = "{}/package_map_data/output".format(cwd)
    pm = PackageMapper(BUILD_TBZ2, SupportedFileType.TarBz2, SupportedFileType.TarBz2, output_dir, package_map)
    pm.apply()
    pm.cleanup()

    expected_output = ['assets.tar.bz2', 'configuration.tar.bz2', 'binaries.tar.bz2', 'libraries.tar.bz2']
    files = os.listdir(output_dir)
    assert set(expected_output) == set([f for f in files if f[0] != '.'])   # filter out .keep

    packages = ['configuration.tar.bz2', 'binaries.tar.bz2', 'libraries.tar.bz2', 'assets.tar.bz2']
    # note trailing / removed from empty directories (per tarfile output)
    expected_members = {
        'configuration.tar.bz2': ['config/file.yaml', 'config/file.xml', 'config/file.txt'],
        'binaries.tar.bz2': ['build-123/bin/lib', 'build-123/bin/widgets', 'build-123/bin/lib/libfoo.so'],
        'libraries.tar.bz2': ['build-123/bin/lib/libfoo.so'],
        'assets.tar.bz2': ['web/static/build-123/assets/css', 'web/static/build-123/assets/css/style.css',
                       'web/static/build-123/assets/main.html', 'web/static/build-123/assets/img',
                       'web/static/build-123/assets/img/banner.png']
    }

    for p in packages:
        pname = "{}/{}".format(output_dir, p)
        with tarfile.open(pname, 'r:bz2') as tf:
            assert set(tf.getnames()) == set(expected_members[p])
        os.unlink(pname)



#from M2Crypto import BIO, RSA
#import base64
import elita.util


__author__ = 'bkeroack'



class KeyPair:
    __metaclass__ = elita.util.LoggingMetaClass

    def __init__(self, private_key, public_key):
        pass
        #self.private_key = BIO.MemoryBuffer(private_key)
        #self.public_key_pkcs1 = public_key
        #self.public_key_x501 = None

    # def der_length(self, length):
    #     if length < 128:
    #         return chr(length)
    #     prefix = 0x80
    #     result = ''
    #     while length > 0:
    #         result = chr(length & 0xff) + result
    #         length >>= 8
    #         prefix += 1
    #     return chr(prefix) + result
    #
    # def process_public_key(self):
    #     '''converts PKCS#1 public key as output by ssh-keygen to X.501'''
    #     pk = self.public_key_pkcs1.split('\n')
    #     pk = '\0' + base64.decodestring("".join(pk[1:-2]))
    #     pk = '\x30\x0d\x06\x09\x2a\x86\x48\x86\xf7\x0d\x01\x01\x01\x05\x00\x03' + self.der_length(len(pk)) + pk
    #     pk = '\x30' + self.der_length(len(pk)) + pk
    #     pk = '-----BEGIN PUBLIC KEY-----\n' + base64.encodestring(pk) + '-----END PUBLIC KEY-----'
    #     self.public_key_x501 = BIO.MemoryBuffer(pk)

    def verify_public(self):
        return  # key verification is a pain in the ass
        #rsa = RSA.load_pub_key_bio(self.public_key_x501)
        #util.debugLog(self, "public key loaded successfully")

    def verify_private(self):
        return
        #rsa = RSA.load_key_bio(self.private_key)
        #util.debugLog(self, "private key loaded successfully")



__author__ = 'bkeroack'

from types import FunctionType
import functools
import logging
import collections

import type_check

def log_wrapper(func):
    @functools.wraps(func)
    def log(*args, **kwargs):
        logging.debug("CALLING: {} (args: {}, kwargs: {}".format(func.__name__, args, kwargs))
        ret = func(*args, **kwargs)
        logging.debug("{} returned: {}".format(func.__name__, ret))
        return ret
    return log


class LoggingMetaClass(type):
    def __new__(mcs, classname, bases, class_dict):
        new_class_dict = dict()
        for attr_name, attr in class_dict.items():
            new_class_dict[attr_name] = log_wrapper(attr) if type(attr) == FunctionType else attr
        return type.__new__(mcs, classname, bases, new_class_dict)


def change_dict_keys(obj, char, rep):
    '''
    Recursively replaces char in nested dict keys with rep (for sanitizing input to mongo, for example)
    Modifies object in place! Returns None.
    '''
    new_obj = obj
    for k in new_obj:
            if isinstance(k, list):
                change_dict_keys(k, char, rep)
            if isinstance(obj, dict):
                if isinstance(obj[k], dict):
                        change_dict_keys(obj[k], char, rep)
                if char in k:
                        obj[k.replace(char, rep)] = obj[k]
                        del obj[k]


def split_seq(seq, n):
    '''split iterable into n number of equal-length (approx) chunks'''
    newseq = []
    splitsize = 1.0/n*len(seq)
    for i in range(n):
        ns = seq[int(round(i*splitsize)):int(round((i+1)*splitsize))]
        if len(ns) > 0:
            newseq.append(ns)
    return newseq

def set_dict_key(obj, path, value):
    '''
    In the dict-like obj (assumed to be a nested set of dicts), walk path and insert value.

    For example,

    obj = root_tree
    path = ('app', 'myapp', 'builds')
    value = { [build_reference_doc] }
    '''
    for k in path[:-1]:
        obj = obj.setdefault(k, {})
    obj[path[-1]] = value

def paths_from_nested_dict(dict_obj, path=None):
    '''
    Given an arbitrarily-nested dict-like object, generate a list of unique tree path tuples.
    The last object in any path will be the deepest leaf value in that path.
    Ex:
    dict_obj = {
        'a': {
            0: 1,
            1: 2
        },
        'b': {
            'foo': 'bar'
        }
    }

    returns:
    [
        ('a', 0, 1),
        ('a', 1, 2),
        ('b', 'foo', 'bar')
    ]

    @type dict_obj: dict
    @type path: list
    '''
    assert not path or hasattr(path, '__getitem__')
    assert type_check.is_dictlike(dict_obj)
    assert not path or isinstance(path, list)
    path = path if path else list()
    unique_paths = list()
    for i, item in enumerate(dict_obj.iteritems()):
        if type_check.is_dictlike(item[1]):
            for nested_item in paths_from_nested_dict(item[1], path=path+[item[0]]):
                unique_paths.append(nested_item)
        else:
            unique_paths.append(tuple(path + [item[0]] + [item[1]]))
    return unique_paths

def flatten(list_obj):
    '''
    Given a nested n-dimensional list, return a flattened list
    '''
    assert list_obj
    assert isinstance(list_obj, collections.Iterable)
    for item in list_obj:
        if isinstance(item, collections.Iterable) and not type_check.is_string(item):
            for x in flatten(item):
                yield x
        else:
            yield item

def flatten_list(list_obj):
    '''
    Above without having to iterate
    '''
    return [x for x in flatten(list_obj)]



import os

from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.rst')).read()
CHANGES = open(os.path.join(here, 'RELEASE_CHANGES.rst')).read()

requires = [
    #core deps
    'pyramid',
    'gunicorn',
    'gevent',
    'salt<=2014.1.10',     # temp fix since SaltStack, Inc. likes to push broken pre-release software to PyPI
    'celery',
    'pymongo',
    'glob2',

    # direct git integration
    'gitpython>=0.3.2.RC1',

    #for api calls, calculate slugs for BitBucket URLs
    'requests',
    'python-slugify',
    'simplejson',

    # to parse salt states
    'PyYAML',

    # to parse ssh config
    #'paramiko',

    # we need to lock, eg, salt state files during modification
    'lockfile',

    # for installation primarily
    'sh',
    'clint',

    #testing
    'nose',
    'mock'
]

setup(name='elita',
      version="0.63.0",
      description='Continuous deployment (continuous delivery) and infrastructure management REST framework',
      long_description=README + '\n\n' + CHANGES,
      license='Apache',
      classifiers=[
        "Programming Language :: Python",
        "Framework :: Pyramid",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: WSGI :: Application",
        "Topic :: System :: Systems Administration",
        "Topic :: Software Development :: Build Tools"
        ],
      author='Benjamen Keroack',
      author_email='ben@elita.io',
      url='https://bitbucket.org/scorebig/elita',
      keywords='continuous deployment delivery REST automation devops',
      packages=find_packages(exclude=['test*']),
      package_data={
          'elita_install': ['util/*', '*.ini', '*.txt', '*.rst', '*.cfg'],
          'elita_cli': ['*']
      },
      include_package_data=True,
      zip_safe=False,
      install_requires=requires,
      entry_points={
          'paste.app_factory': ['main=elita:main'],
          'console_scripts': [
              'elita_install=elita_install:Install',
              'elita=elita_cli:Command_Line_Client'
          ]
      })



import salt.client
import salt.config
from salt.exceptions import SaltClientError
import lockfile
import os
import os.path
import random
import shutil
import string
import yaml
import logging
import copy
import elita.deployment.gitservice

try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import Loader, Dumper

import elita.util
import elita.util.type_check

__author__ = 'bkeroack'

class FatalSaltError(Exception):
    pass

class OSTypes:
    Windows = 0
    Unix_like = 1

class RemoteCommands:
    __metaclass__ = elita.util.LoggingMetaClass

    def __init__(self, sc):
        '''
        @type sc: SaltController
        '''
        assert sc
        assert isinstance(sc, SaltController)
        self.sc = sc

    def get_os(self, server):
        resp = self.sc.salt_command(server, 'grains.item', ["os"])
        logging.debug("get_os: resp: {}".format(resp))
        return OSTypes.Windows if resp[server]['os'] == "Windows" else OSTypes.Unix_like

    def get_os_text(self, server):
        ost = self.get_os(server)
        if ost == OSTypes.Windows:
            return "windows"
        elif ost == OSTypes.Unix_like:
            return "unix"
        else:
            return "unknown"

    def group_remote_servers_by_os(self, server_list):
        '''
        Determine which servers are Windows and which are Unix-like, so we know where to push keys, etc to.
        '''
        unix_servers = list()
        win_servers = list()
        for s in server_list:
            ost = self.get_os(s)
            if ost == OSTypes.Windows:
                win_servers.append(s)
            elif ost == OSTypes.Unix_like:
                unix_servers.append(s)
        return {
            'unix': unix_servers,
            'windows': win_servers
        }

    def get_remote_home(self, server_list):
        '''
        Get shell expansion of ~ on remote servers (home of user salt-minion is running as)

        @type server_list: list(str)
        '''
        assert server_list
        assert elita.util.type_check.is_seq(server_list)
        servers_by_os = self.group_remote_servers_by_os(server_list)
        res_win = self.sc.salt_command(servers_by_os['windows'], 'cmd.run', ['$env:userprofile'],
                                        opts={'shell': 'powershell'}) if len(servers_by_os['windows']) > 0 else {}
        res_unix = self.sc.salt_command(servers_by_os['unix'], 'cmd.run', ['eval "echo ~$(whoami)"']) \
            if len(servers_by_os['unix']) > 0 else {}
        assert len(set(res_unix.keys()).intersection(set(res_win.keys()))) == 0     # no common keys
        return dict(res_unix, **res_win)

    def get_home_expanded_path(self, server_list, path_list):
        '''
        Given a list of servers and a list of paths with '~' in them, return absolute paths

        @type server_list: list(str)
        @type path_list: list(str)
        '''

        assert server_list and path_list
        assert elita.util.type_check.is_seq(server_list)
        assert elita.util.type_check.is_seq(path_list)
        assert all(['~' in p for p in path_list])

        path_by_server = {item[0]: item[1] for item in zip(server_list, path_list)}
        home_dirs = self.get_remote_home(server_list)
        return {s: path_by_server[s].replace('~', home_dirs[s]) for s in path_by_server}

    def create_directory(self, server_list, path):
        assert isinstance(server_list, list)
        return self.sc.salt_command(server_list, 'file.mkdir', [path])

    def delete_directory_win(self, server_list, path):
        return self.sc.salt_command(server_list, 'cmd.run', ["if (Test-Path {path}) {{ Remove-Item -Recurse -Force {path} }}".format(path=path)],
                                    opts={'shell': 'powershell'})

    def delete_directory_unix(self, server_list, path):
        return self.sc.salt_command(server_list, 'cmd.run', ["rm -rf {}".format(path)])

    def delete_directory(self, server_list, path):
        assert isinstance(server_list, list)
        res = list()
        win_s = list()
        unix_s = list()
        for s in server_list:
            ost = self.get_os(s)
            if ost == OSTypes.Windows:
                win_s.append(s)
            elif os == OSTypes.Unix_like:
                unix_s.append(s)
        if len(win_s) > 0:
            res.append(self.delete_directory_win(win_s, path))
        if len(unix_s) > 0:
            res.append(self.delete_directory_unix(unix_s, path))
        return res

    def _get_salt_uri(self, local_path):
        '''
        For pushing local files to remote servers, need to copy it to a random location in salt tree
        so we can push via cp.get_file. This file should be deleted after that call returns.
        '''

        r_postfix = ''.join(random.sample(string.letters + string.digits, 20))
        root = self.sc.sls_dir
        newname = os.path.basename(local_path) + r_postfix
        new_fullpath = "{}/{}".format(root, newname)
        shutil.copy(local_path, new_fullpath)
        return 'salt://{}/{}'.format(self.sc.settings['elita.salt.slsdir'], newname), new_fullpath

    def push_files(self, target, local_remote_path_mapping, mode=None):
        '''
        Push a list of files to corresponding remote paths. Make sure each remote path exists first.
        Do all pushes in a single salt call (faster).

        Note that these must be files, not directories.

        local_remote_path_mapping = { 'local_path': 'remote_path' }
        '''
        assert target and local_remote_path_mapping
        assert elita.util.type_check.is_seq(target)
        assert elita.util.type_check.is_dictlike(local_remote_path_mapping)
        # paths can't be directories
        assert all([p[-1] != '/' and local_remote_path_mapping[p][-1] != '/' for p in local_remote_path_mapping])
        assert elita.util.type_check.is_optional_str(mode)

        calls = list()
        params = list()
        temp_files = list()

        #delete remote path first
        logging.debug("removing any existing remote_path")
        for local_path in local_remote_path_mapping:
            self.rm_dir_if_exists(target, local_remote_path_mapping[local_path])

        for local_path in local_remote_path_mapping:
            if 'salt://' in local_path:
                salt_uri = local_path
            else:
                salt_uri, tf = self._get_salt_uri(local_path)
                temp_files.append(tf)
            remote_path = local_remote_path_mapping[local_path]
            expanded_path = self.get_home_expanded_path(target, [remote_path])[target[0]] if '~' in remote_path else remote_path
            calls.append('cp.get_file')
            if mode:
                calls.append('file.set_mode')
            params.append([salt_uri, expanded_path])
            if mode:
                params.append([expanded_path, mode])
            logging.debug("push_files: remote_path: {}".format(expanded_path))

        rets = self.sc.salt_command(target, calls, params, opts={'makedirs': True})
        for tf in temp_files:
            os.unlink(tf)
        return rets

    def rm_file_if_exists(self, target, path):
        '''
        Remove target directory if it exists
        '''
        assert target and path
        assert elita.util.type_check.is_seq(target)
        assert elita.util.type_check.is_string(path)
        for server in target:
            expanded_path = self.get_home_expanded_path([server], [path])[server] if '~' in path else path
            file_exists = self.sc.salt_command([server], "file.file_exists", [expanded_path])
            if server in file_exists and file_exists[server]:
                self.sc.salt_command([server], "file.remove", [expanded_path])

    def rm_dir_if_exists(self, target, path):
        '''
        Remove target directory if it exists
        '''
        assert target and path
        assert elita.util.type_check.is_seq(target)
        assert elita.util.type_check.is_string(path)
        servers_by_os = self.group_remote_servers_by_os(target)
        for server in target:
            expanded_path = self.get_home_expanded_path([server], [path])[server] if '~' in path else path
            file_exists = self.sc.salt_command([server], "file.file_exists", [expanded_path])
            if file_exists[server]:
                self.sc.salt_command([server], "file.remove", [expanded_path])
                break
            dir_exists = self.sc.salt_command([server], "file.directory_exists", [expanded_path])
            if dir_exists[server]:
                #have to delete recursively, and no built-in salt module supports that
                if server in servers_by_os['unix']:
                    self.sc.salt_command([server], 'cmd.run', ['rm -rf {}'.format(expanded_path)])
                elif server in servers_by_os['windows']:
                    self.sc.salt_command([server], 'cmd.run', ['Remove-Item -Recurse -Force {}'.format(expanded_path)],
                                         opts={'shell': 'powershell'})
                else:
                    raise FatalSaltError

    def run_powershell_script(self, server_list, script_path):
        return self.sc.salt_command(server_list, 'cmd.run', [script_path], opts={'shell': 'powershell'})

    def run_shell_command(self, server_list, cmd):
        return self.sc.salt_command(server_list, 'cmd.run', [cmd])

    def touch(self, server_list, remote_path):
        #hack: assuming all servers in server_list have the same ~ path expansion
        expanded_path = self.get_home_expanded_path(server_list, [remote_path])[server_list[0]] if '~' in remote_path else remote_path
        return self.sc.salt_command(server_list, 'file.touch', [expanded_path])

    def ping(self, server_list, timeout=10):
        return self.sc.salt_command(server_list, 'test.ping', arg=[], timeout=timeout)

    def append_to_file(self, server_list, remote_path, text_block):
        #hack: assuming all servers in server_list have the same ~ path expansion
        expanded_path = self.get_home_expanded_path(server_list, [remote_path])[server_list[0]] if '~' in remote_path else remote_path
        return self.sc.salt_command(server_list, 'file.append', [expanded_path, text_block])

    def clone_repo(self, server_list, repo_uri, dest):
        assert server_list and repo_uri and dest
        assert elita.util.type_check.is_seq(server_list)
        assert elita.util.type_check.is_string(repo_uri)
        assert elita.util.type_check.is_string(dest)
        logging.debug("clone_repo: deleting remote destination if necessary")
        self.rm_dir_if_exists(server_list, dest)
        cwd, dest_dir = os.path.split(dest)
        return self.sc.salt_command(server_list, 'cmd.run',
                                    ['git clone {uri} {dest}'.format(uri=repo_uri, dest=dest_dir)],
                                    opts={'cwd': cwd}, timeout=1200)

    def checkout_branch(self, server_list, location, branch_name):
        return self.sc.salt_command(server_list, 'cmd.run', ['git checkout {}'.format(branch_name)],
                                    opts={'cwd': location})

    def discard_git_changes(self, server_list, location):
        return self.sc.salt_command(server_list, 'cmd.run', ['git checkout -f'],
                                    opts={'cwd': location})

    def add_all_files_git(self, server_list, location):
        return self.sc.salt_command(server_list, 'cmd.run', ['git add -A'], opts={'cwd': location}, timeout=120)

    def commit_git(self, server_list, location, message):
        return self.sc.salt_command(server_list, 'cmd.run', ['git commit -m "{}"'.format(message)],
                                    opts={'cwd': location}, timeout=30)

    def set_user_git(self, server_list, location, email, name):
        return {
            "name": self.sc.salt_command(server_list, 'cmd.run', ['git config user.name "{}"'.format(name)],
                                         opts={'cwd': location}, timeout=30),
            "email": self.sc.salt_command(server_list, 'cmd.run', ['git config user.email "{}"'.format(email)],
                                         opts={'cwd': location}, timeout=30)
        }

    def set_git_autocrlf(self, server_list, location):
        return self.sc.salt_command(server_list, 'cmd.run', ['git config core.autocrlf input'],
                                    opts={'cwd': location}, timeout=30)

    def set_git_push_url(self, server_list, location, url):
        return self.sc.salt_command(server_list, 'cmd.run', ['git remote set-url --push origin "{}"'.format(url)],
                                    opts={'cwd': location}, timeout=30)

    # def set_git_upstream(self, server_list, location, branch):
    #     return self.sc.salt_command(server_list, 'cmd.run',
    #                                 ['git branch --set-upstream-to=origin/{} {}'.format(branch, branch)],
    #                                 opts={'cwd': location}, timeout=30)

    def highstate(self, server_list):
        return self.sc.salt_command(server_list, 'state.highstate', [], timeout=300)

    def run_sls(self, server_list, sls_name):
        return self.sc.salt_command(server_list, 'state.sls', [sls_name], timeout=300)

    def run_slses_async(self, callback, sls_map, args=None):
        '''
        sls_map: { 'sls_name': [ list_of_servers ] }
        Runs all listed SLSes asynchronously
        '''
        cmd_group = [{'command': 'state.sls', 'arguments': [s], 'servers': sls_map[s]} for s in sls_map]
        return self.sc.salt_commands_async(callback, args, cmd_group, timeout=600, opts={'concurrent': True})

class SaltController:
    __metaclass__ = elita.util.LoggingMetaClass

    def __init__(self, datasvc):
        '''
        @type datasvc: elita.models.DataService
        '''
        self.settings = datasvc.settings
        self.salt_client = salt.client.LocalClient()
        self.datasvc = datasvc
        self.file_roots = None
        self.pillar_roots = None
        self.sls_dir = None

        try:
            self.load_salt_info()
        except SaltClientError:
            logging.debug("WARNING: SaltClientError")

    def salt_commands_async(self, callback, args, cmd_group, opts=None, timeout=120):
        rets = list()
        for c in cmd_group:
            rets.append((c, self.salt_client.cmd_iter(c['servers'], c['command'], c['arguments'], kwarg=opts,
                                      timeout=timeout, expr_form='list')))
        results = list()
        for retc in rets:
            for r in retc[1]:
                callback(r, retc[0]['arguments'], **args)
                results.append(r)
        return results

    def salt_command(self, target, cmd, arg, opts=None, timeout=20):
        results = self.salt_client.cmd(target, cmd, arg, kwarg=opts, timeout=timeout, expr_form='list')
        results_sanitized = copy.deepcopy(results)
        elita.util.change_dict_keys(results_sanitized, '.', '_')    # remove '.' in keys for mongo
        # try to detect if there was an error or exception in any results
        for path in elita.util.paths_from_nested_dict(results):
            if elita.util.type_check.is_string(path[-1]):
                if not all([word.lower() not in path[-1].lower() for word in ('error', 'exception', 'traceback', 'fatal')]):
                    self.datasvc.jobsvc.NewJobData({'error': 'error detected in salt command results',
                                                    'detected_in': path[-1],
                                                    'all_results': results_sanitized,
                                                    'target': target,
                                                    'cmd': cmd,
                                                    'arg': arg})
                    raise FatalSaltError
        self.datasvc.jobsvc.NewJobData({'salt_command': {'target': target, 'cmd': cmd, 'arg': arg,
                                                         'results': results_sanitized}})
        return results

    def run_command(self, target, cmd, shell, timeout=120):
        return self.salt_command(target, 'cmd.run', [cmd], opts={"shell": shell}, timeout=timeout)

    def load_salt_info(self):
        master_config = salt.config.master_config(os.environ.get('SALT_MASTER_CONFIG', '/etc/salt/master'))
        self.file_roots = master_config['file_roots']
        for e in self.file_roots:
            for d in self.file_roots[e]:
                if not os.path.isdir(d):
                    os.mkdir(d)
        self.pillar_roots = master_config['pillar_roots']
        for e in self.pillar_roots:
            for d in self.pillar_roots[e]:
                if not os.path.isdir(d):
                    os.mkdir(d)
        self.sls_dir = "{}/{}".format(self.file_roots['base'][0], self.settings['elita.salt.slsdir'])
        if not os.path.isdir(self.sls_dir):
            os.mkdir(self.sls_dir)

    def verify_connectivity(self, server, timeout=10):
        return len(self.salt_client.cmd(server, 'test.ping', timeout=timeout)) != 0

    def get_gd_file_name(self, app, name):
        if not os.path.isdir(self.sls_dir):
            os.mkdir(self.sls_dir)
        appdir = "{}/{}".format(self.sls_dir, app)
        if not os.path.isdir(appdir):
            os.mkdir(appdir)
        return "{}/{}.sls".format(appdir, name)

    def get_gitdeploy_entry_name(self, app, gd_name):
         return "{}.{}.{}".format(self.settings['elita.salt.slsdir'], app, gd_name)


    def rm_gitdeploy_yaml(self, gitdeploy):
        name = gitdeploy['name']
        app = gitdeploy['application']
        filename = self.get_gd_file_name(app, name)
        logging.debug("rm_gitdeploy_yaml: gitdeploy: {}/{}".format(app, name))
        logging.debug("rm_gitdeploy_yaml: filename: {}".format(filename))
        if not os.path.isfile(filename):
            logging.debug("rm_gitdeploy_yaml: WARNING: file not found!")
            return
        lock = lockfile.FileLock(filename)
        logging.debug("rm_gitdeploy_yaml: acquiring lock on {}".format(filename))
        lock.acquire(timeout=60)
        os.unlink(filename)
        lock.release()

    def new_gitdeploy_yaml(self, gitdeploy):
        '''Adds new gitdeploy to existing YAML file or creates new YAML file.'''
        name = gitdeploy['name']
        app = gitdeploy['application']
        filename = self.get_gd_file_name(app, name)
        logging.debug("add_gitdeploy_to_yaml: acquiring lock on {}".format(filename))
        lock = lockfile.FileLock(filename)
        lock.acquire(timeout=60)  # throws exception if timeout
        if os.path.isfile(filename):
            logging.debug("add_gitdeploy_to_yaml: existing yaml sls")
            with open(filename, 'r') as f:
                existing = yaml.load(f, Loader=Loader)
            if not existing:    # if empty file this will be None
                existing = dict()
        else:
            logging.debug("add_gitdeploy_to_yaml: yaml sls does not exist")
            existing = dict()
        slsname = 'gitdeploy_{}'.format(name)
        favor = "theirs" if gitdeploy['options']['favor'] in ('theirs', 'remote') else 'ours'
        ignore_whitespace_op = "-Xignore-all-space" if gitdeploy['options']['ignore-whitespace'] == 'true' else ""
        dir = gitdeploy['location']['path']
        existing[slsname] = {
            'cmd.run': [
                {
                    'name': 'git pull -s recursive -X{favor} -Xpatience {ignorews}'
                    .format(favor=favor, ignorews=ignore_whitespace_op)
                },
                {
                    'cwd': dir
                },
                {
                    'failhard': 'true'
                },
                {
                    'order': 1
                }
            ]
        }
        prepull = gitdeploy['actions']['prepull']
        if prepull is not None:
            elita.util.change_dict_keys(prepull, elita.deployment.gitservice.EMBEDDED_YAML_DOT_REPLACEMENT, '.')
            for k in prepull:
                top_key = prepull[k].keys()[0]
                prepull[k][top_key].append({'order': 0})
                existing[k] = prepull[k]
        postpull = gitdeploy['actions']['postpull']
        if postpull is not None:
            elita.util.change_dict_keys(postpull, elita.deployment.gitservice.EMBEDDED_YAML_DOT_REPLACEMENT, '.')
            for k in postpull:
                top_key = postpull[k].keys()[0]
                postpull[k][top_key].append({'order': 'last'})
                existing[k] = postpull[k]
        with open(filename, 'w') as f:
            f.write(yaml.safe_dump(existing, default_flow_style=False))
        lock.release()





import pymongo
import sys
import traceback
import logging

import elita.util
import elita.models
import elita.celeryinit

__author__ = 'bkeroack'

def regen_datasvc(settings, job_id):
    client = pymongo.MongoClient(settings['elita.mongo.host'], int(settings['elita.mongo.port']))
    db = client[settings['elita.mongo.db']]
    tree = db['root_tree'].find_one()
    updater = elita.models.RootTreeUpdater(tree, db)
    root = elita.models.RootTree(db, updater, tree, None)
    return client, elita.models.DataService(settings, db, root, job_id=job_id)

@elita.celeryinit.celery.task(bind=True, name="elita_task_run_job")
def run_job(self, settings, callable, args):
    ''' Generate new dataservice, run callable, store results
    '''
    job_id = self.request.id
    client, datasvc = regen_datasvc(settings, job_id)
    try:
        results = callable(datasvc, **args)
    except:
        exc_type, exc_obj, tb = sys.exc_info()
        f_exc = traceback.format_exception(exc_type, exc_obj, tb)
        results = {
            "error": "unhandled exception during callable!",
            "exception": f_exc
        }
        logging.debug("EXCEPTION: {}".format(f_exc))
    datasvc.jobsvc.SaveJobResults(results if results else {"status": "job returned no data"})
    client.close()

#generic interface to run code async (not explicit named actions/hooks)
def run_async(datasvc, name, job_type, data, callable, args):
    logging.debug("run_async: create new async task: {}; args: {}".format(callable, args))
    job = datasvc.jobsvc.NewJob(name, job_type, data)
    job_id = str(job.job_id)
    run_job.apply_async((datasvc.settings, callable, args), task_id=job_id)
    return job_id

class ActionService:
    __metaclass__ = elita.util.LoggingMetaClass

    def __init__(self, datasvc):
        self.datasvc = datasvc

    def register(self):
        self.hooks = RegisterHooks(self.datasvc)
        self.hooks.register()
        self.actions = RegisterActions(self.datasvc)
        self.actions.register()

    def get_action_details(self, app, action_name):
        return self.actions.actionmap[app][action_name] if app in self.actions.actionmap and action_name in self\
            .actions.actionmap[app] else None

    def async(self, app, action_name, params):
        action = self.actions.actionmap[app][action_name]['callable']
        job = self.datasvc.jobsvc.NewJob("{}.{}".format(app, action_name), "Action", {
            "application": app,
            "params": params
        })
        job_id = str(job.job_id)
        run_job.apply_async((self.datasvc.settings, action, {'params': params}), task_id=job_id)
        return {"action": action_name, "job_id": job_id, "status": "async/running"}


DefaultHookMap = {
    'BUILD_UPLOAD_SUCCESS': None,
    'GITDEPLOY_INIT_PRE': None,
    'GITDEPLOY_INIT_POST': None,
    'GITDEPLOY_DEINIT_PRE': None,
    'GITDEPLOY_DEINIT_POST': None,
    'GITDEPLOY_COMMIT_DIFF': None,
    'AUTO_DEPLOYMENT_START': None,
    'AUTO_DEPLOYMENT_BATCH_BEGIN': None,
    'AUTO_DEPLOYMENT_BATCH_DONE': None,
    'AUTO_DEPLOYMENT_COMPLETE': None,
    'AUTO_DEPLOYMENT_FAILED': None
}

class RegisterHooks:
    __metaclass__ = elita.util.LoggingMetaClass

    def __init__(self, datasvc):
        self.datasvc = datasvc
        self.hookmap = dict()

    def register(self):
        #first load all defaults
        apps = self.datasvc.appsvc.GetApplications()
        for app in apps:
            self.hookmap[app] = {k: DefaultHookMap[k] for k in DefaultHookMap}  # need a new dict for each app
        from pkg_resources import iter_entry_points
        for obj in iter_entry_points(group="elita.modules", name="register_hooks"):
            hooks = (obj.load())()  # returns: { app: { "HOOK_NAME": <callable> } }
            for app in hooks:
                if app in apps:
                    for a in hooks[app]:
                        self.hookmap[app][a] = hooks[app][a]
                else:
                    logging.debug("register: WARNING: unknown application '{}'".format(app))
            logging.debug("HookMap: {}".format(self.hookmap))

    def run_hook(self, app, name, args):
        # Hooks are always triggered in an async context, so there's no need to spawn another celery task
        hook = self.hookmap[app][name]
        if hook is None:
            return "none"
        args['datasvc'] = self.datasvc
        return hook(**args)


class RegisterActions:
    __metaclass__ = elita.util.LoggingMetaClass

    def __init__(self, datasvc):
        self.datasvc = datasvc
        self.actionmap = dict()

    def register(self):
        logging.debug("register")

        from pkg_resources import iter_entry_points
        apps = self.datasvc.appsvc.GetApplications()

        for obj in iter_entry_points(group="elita.modules", name="register_actions"):
            logging.debug("register: found obj: {}".format(obj))
            actions = (obj.load())()
            logging.debug("actions: {}".format(actions))
            for app in actions:
                if app not in apps:
                    logging.debug("WARNING: application not found: {}".format(app))
                else:
                    if app not in self.actionmap:
                        self.actionmap[app] = dict()
                    for a in actions[app]:
                        action_name = a['callable'].__name__
                        self.actionmap[app][action_name] = a
                        logging.debug("NewAction: app: {}; action_name: {}; params: {}".format(app, action_name, a['params']))
                        params = a['params'].keys()
                        self.datasvc.jobsvc.NewAction(app, action_name, params)
